#ifndef __ANSITERM_H__
#define __ANSITERM_H__
#define MAX_ARGS 15
#define ESC 27

#define ATTR_BOLD 1
#define ATTR_REVERSE 2
#define ATTR_BLINK 4
#define ATTR_UNDER 8
#define ATTR_SELECTION 16

#define DEBUG 
/* g_print("%s(%s)\n", __FILE__, __FUNCTION__)  */

#define linenumber(l) (((l) + term->visible_line_begin)%term->saved_lines)

#define CUR_APL 1
#define EDITOR_BUTTON 2


typedef struct _AnsiTerm AnsiTerm;
typedef struct _AnsiTermClass AnsiTermClass;

typedef int (*ToTermFunc)(AnsiTerm *, char *buf, int nbytes);
typedef int (*FromTermFunc)(char *buf, int nbytes);

typedef struct _KeyMap KeyMap;
typedef void (*KeyMapFunc)(gpointer);
struct _KeyMap {
	int keyval;
	int state;
	gpointer data;
	KeyMapFunc func;
};

struct  _AnsiTerm {
	GtkWidget widget;
	GtkAdjustment *adjustment;
	GtkWidget *scrollbar;
	GdkGC *gc;
	GdkFont *font;
	GdkGC *white_gc;
	GdkGC *black_gc;
	GdkGC *bg_gc;
	GdkGC *cursor_gc;
	guchar *buf;
	guchar *attr;
	guchar *color;
	guchar state;
	guchar cu_fg, old_fg;
	guchar cu_bg, old_bg;
	guchar normal_fg, normal_bg;
	guchar cu_attr, old_attr;
	guchar cu_x, old_x;
	gchar cu_y, old_y;
	guchar top, bot;
	guchar col, row;
	guchar argc;
	guchar max_reached;
	guint16 saved_lines;
	guint16 visible_line_begin;
	guint16 real_visible_line_begin;
	guint16 argv[MAX_ARGS];
	guchar font_height;
	guchar font_width;
	guchar font_ascent;
	gboolean use_bold, use_under, auto_linefeed, silent, draw_flag;
	gboolean can_block;
	ToTermFunc to_term;
	FromTermFunc from_term;
	GList *keymap_list;
	char temp_hangul[5];
	char charset;
	int mode;
	char *tmp_str;
	int tmp_alloc, tmp_len;
};

struct _AnsiTermClass {
	GtkWidgetClass parent_class;
	void (*size_changed) (AnsiTerm *term);
	void (*line_changed) (AnsiTerm *term);
};

#define ANSI_TERM(w) ((AnsiTerm *)w)
GtkWidget * ansi_term_new(char *terminal_font, int col, int row, 
	int saved_lines, int font_ascent, int font_height, 
	ToTermFunc to_term, FromTermFunc from_term);
void keymap_init(AnsiTerm *);
void show_temp_hangul(AnsiTerm *);
void expose_line(AnsiTerm *term, int y, int x1, int x2);
void ansi_term_show_cursor(AnsiTerm *term);
void ansi_term_hide_cursor(AnsiTerm *term);
void ansi_term_reset(AnsiTerm *term);
void ansi_term_flush(AnsiTerm *term);
gint ansi_term_key_press(GtkWidget *w, GdkEventKey *ev);
int ansi_term_selection_clear(GtkWidget *w, GdkEventSelection *ev);
int ansi_term_selection_request(GtkWidget *w, GdkEventSelection *ev);
int ansi_term_selection_notify(GtkWidget *w, GdkEventSelection *ev);
int button_press_callback(GtkWidget *w, GdkEventButton *ev);
int button_release_callback(GtkWidget *w, GdkEventButton *ev);
int motion_callback(AnsiTerm *term, GdkEventMotion *ev);
void check_in_selection(AnsiTerm *term);


#endif /*  __ANSITERM_H__ */
