
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <ctype.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "gau.h"
#include "gauchat.h"
#include "gauscrolledchat.h"
#include "config.h"

GtkWidget *chat_win = NULL;
GauScrolledChat *schat = NULL;
char *chat_prefix1 = NULL;
char *chat_prefix2 = NULL;
char *chat_suffix = NULL;
int chat_append = FALSE;
static void flush_callback(GauChat *chat, char *buf, int len);
static void chat_add_choose_callback(GtkWidget *w, GauChat *chat);
static void chat_add_head_choose_callback(GtkWidget *w, GauChat *chat);
static void chat_add_callback(GtkWidget *w, GauChat *chat);
static int chat_ctrl_z(GauChat *chat, GdkEventKey *ev, AnsiTerm *term);

void
window_destroyed(GtkWidget *w, GtkWidget **win)
{
	*win = NULL;
}

static void
flush_callback(GauChat *chat, char *buf, int len)
{
	term->from_term(buf, len);
	term->from_term("
", 1);
}

void chat_binding_prepare(GauChat *chat);

void
chat_callback(GtkWidget *w, gpointer data)
{
	int x, y;
	int width, height;
	GdkGCValues gcvalue;
	GtkWidget *vbox, *hbox, *button;
	gdk_window_get_origin(GTK_WIDGET(term)->window, &x, &y);
	gdk_window_get_size (GTK_WIDGET(term)->window, &width, &height);

	y += height + 4;

	if(chat_win != NULL) {
	    if(GTK_WIDGET_VISIBLE(chat_win)) {
		gtk_widget_hide(chat_win);
	    } else
		gtk_widget_popup(chat_win, x, y);
	    return;
	}
	chat_win = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_set_policy(GTK_WINDOW(chat_win), FALSE, FALSE, FALSE);
	//gtk_widget_set_usize(win, width, -1);
	gtk_window_set_title(GTK_WINDOW(chat_win), ": ̾߱ â");
	gtk_signal_connect(GTK_OBJECT(chat_win), "destroy",
		(GtkSignalFunc)window_destroyed, &chat_win);
	gtk_container_border_width(GTK_CONTAINER(chat_win), 5);

	hbox = gtk_hbox_new(FALSE, 4);
	gtk_container_add(GTK_CONTAINER(chat_win), hbox);
	gtk_widget_show(hbox);

	gdk_gc_get_values(term->cursor_gc, &gcvalue);
	schat = gau_scrolled_chat_new(gcvalue.font, term->font_height, 
		term->font_ascent, term->col-5, 30, 5, 
		(gpointer)flush_callback);
	if(chat_prefix1) 
		gau_chat_prefix_set(schat->chat, chat_prefix1, 1);
	if(chat_prefix2) 
		gau_chat_prefix_set(schat->chat, chat_prefix2, 2);
	if(chat_suffix)
		gau_chat_suffix_set(schat->chat, chat_suffix);
	gtk_widget_show(schat->win);
	gtk_container_add(GTK_CONTAINER(hbox), schat->win);
	gtk_widget_popup(chat_win, x, y);
	gtk_widget_grab_focus(GTK_WIDGET(schat->chat));
	gtk_signal_connect(GTK_OBJECT(schat->chat), "key_press_event",
		GTK_SIGNAL_FUNC(chat_ctrl_z), term);
	chat_binding_prepare(schat->chat);
	if(!chat_append) return;

	gtk_widget_new(gtk_vseparator_get_type(),
		"GtkWidget::parent", hbox,
		"GtkWidget::visible", TRUE,
		NULL);
	vbox = gtk_vbox_new(TRUE, 5);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, FALSE, 2);
	gtk_widget_show(vbox);

	button = gtk_button_new_with_label(" ");
	GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(chat_add_callback), schat->chat);
	gtk_widget_show(button);

	button = gtk_button_new_with_label(" ");
	GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(chat_add_choose_callback), schat->chat);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("Ӹ1");
	GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(vbox), button, FALSE, FALSE, 2);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(chat_add_head_choose_callback), schat->chat);
	gtk_widget_show(button);

}

void
chat_binding_prepare(GauChat *chat)
{
	static char *bindstring[20] = {0};
	static int init = TRUE;
	FILE *fp;
	char *buf, filename[256];
	int i, index;
	if(init) {
		init = FALSE;
		sprintf(filename, "%s/chat_binding", expand_path(gau_path));
		fp = fopen(filename, "r");
		if(fp == NULL) return;
		while((buf = get_line(fp)) != NULL) {
			buf++;
			if(!isdigit(buf[0])) {
				g_warning("parse error in %s\n", filename);
				continue;
			}
			index = *buf++ - '0';
			if(isdigit(*buf)) {
				index = 10*index + *buf++ - '0';
			}
			index--;
			transform_string(strchr(buf + 1, '"') + 1, &bindstring[index]);
		}
	}
	for(i=0;i<20;i++) {
		if(bindstring[i]) chat->fbinding[i] = bindstring[i];
	}
}

void
choose_cancel_callback(GtkWidget *w, gpointer data)
{
	gtk_widget_destroy(gtk_widget_get_toplevel(w));
}

void
choose_ok_callback(GtkWidget *w, GauChat *chat)
{
	GtkEntry *entry_prefix, *entry_suffix;
	entry_prefix = gtk_object_get_data(GTK_OBJECT(w), "entry_prefix1");
	gau_chat_prefix_set(chat, gtk_entry_get_text(entry_prefix), 1);
	entry_prefix = gtk_object_get_data(GTK_OBJECT(w), "entry_prefix2");
	gau_chat_prefix_set(chat, gtk_entry_get_text(entry_prefix), 2);
	entry_suffix = gtk_object_get_data(GTK_OBJECT(w), "entry_suffix");
	gau_chat_suffix_set(chat, gtk_entry_get_text(entry_suffix));
	gtk_widget_destroy(gtk_widget_get_toplevel(w));
}

static void
chat_add_head_choose_callback(GtkWidget *w, GauChat *chat)
{
	GtkLabel *label = GTK_LABEL(GTK_BUTTON(w)->child);
	if(chat->add == -1) chat->add = -2;
	else if(chat->add == -2) chat->add = -1;
	else if(chat->add == 1) chat->add = 2;
	else if(chat->add == 2) chat->add = 1;
	if(chat->add == -1 || chat->add == 1) {
		gtk_label_set(label, "Ӹ1");
	} else {
		gtk_label_set(label, "Ӹ2");
	}
}

static void
chat_add_choose_callback(GtkWidget *w, GauChat *chat)
{
	static GtkWidget *win;
	GtkWidget *button;

	GtkWidget *hbox, *vbox, *label;
	GtkWidget *entry_prefix1, *entry_prefix2, *entry_suffix;

	win = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER);
	vbox = gtk_vbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(win), vbox);
	gtk_widget_show(vbox);

	hbox = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
	gtk_widget_show(hbox);

	label = gtk_label_new("Ӹ1");
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 2);
	gtk_widget_show(label);

	entry_prefix1 = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(entry_prefix1), chat->prefix1);
	gtk_box_pack_end(GTK_BOX(hbox), entry_prefix1, FALSE, FALSE, 2);
	gtk_widget_show(entry_prefix1);

	hbox = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
	gtk_widget_show(hbox);

	label = gtk_label_new("Ӹ2");
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 2);
	gtk_widget_show(label);

	entry_prefix2 = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(entry_prefix2), chat->prefix2);
	gtk_box_pack_end(GTK_BOX(hbox), entry_prefix2, FALSE, FALSE, 2);
	gtk_widget_show(entry_prefix2);

	hbox = gtk_hbox_new(FALSE, 2);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
	gtk_widget_show(hbox);
	label = gtk_label_new("");
	gtk_misc_set_alignment(GTK_MISC(label), 1.0, 0.5);
	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 2);
	gtk_widget_show(label);

	entry_suffix = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(entry_suffix), chat->suffix);
	gtk_box_pack_end(GTK_BOX(hbox), entry_suffix, FALSE, FALSE, 2);
	gtk_widget_show(entry_suffix);

	gtk_widget_new(gtk_hseparator_get_type(),
		"GtkWidget::parent", vbox,
		"GtkWidget::visible", TRUE,
		NULL);
	hbox = gtk_hbox_new(TRUE, 5);
	gtk_box_pack_end(GTK_BOX(vbox), hbox, FALSE, FALSE, 2);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("OK");
	gtk_object_set_data(GTK_OBJECT(button), "entry_prefix1", entry_prefix1);
	gtk_object_set_data(GTK_OBJECT(button), "entry_prefix2", entry_prefix2);
	gtk_object_set_data(GTK_OBJECT(button), "entry_suffix", entry_suffix);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(choose_ok_callback), chat);
	gtk_box_pack_end(GTK_BOX(hbox), button,
		FALSE, FALSE, 2);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("");
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(choose_cancel_callback), win);
	gtk_box_pack_end(GTK_BOX(hbox), button,
		FALSE, FALSE, 2);
	gtk_widget_show(button);
	gtk_widget_show(win);
}

static void
chat_add_callback(GtkWidget *w, GauChat *chat)
{
	chat->add = - chat->add;
	if(chat->add > 0) {
		gtk_label_set(GTK_LABEL(GTK_BUTTON(w)->child), " Ŀ");
	} else {
		gtk_label_set(GTK_LABEL(GTK_BUTTON(w)->child), " ");
	}
}

static int
chat_ctrl_z(GauChat *chat, GdkEventKey *ev, AnsiTerm *term)
{
	if(ev->keyval == 'z' && (ev->state & GDK_CONTROL_MASK) &&  chat->current_x == 0) {
		term->from_term("", 1);
		return TRUE;
	} else if(ev->keyval == 'c' && (ev->state & GDK_MOD1_MASK)) {
		gtk_widget_hide(gtk_widget_get_toplevel(GTK_WIDGET(chat)));
		return TRUE;
	}
	return FALSE;
}
