
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <sys/types.h>
#include <sys/stat.h>
#include <wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>

#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "mymenubar.h"

#include "phone.h"
#include "ansiterm.h"
#include "modem.h"
#include "config.h"
#include "gau.h"

typedef struct _Command Command;
struct _Command {
	char *label;
	void (*func)(GtkWidget *w, gpointer data);
	gpointer data;
};

static pid_t editor_pid;
static pid_t shell_pid;
GtkWidget *taegeuk;
GtkWidget * quit_button, *call_button;
GtkWidget * capture_button, *hangup_after_transfer_button;
GtkWidget * cancel_button , *chat_button;
GtkWidget *write_button;

void hangup_callback(GtkWidget *w, gpointer data);
void capture_callback(GtkWidget *w, gpointer data);
void exit_callback(GtkWidget *w, gpointer data);
void exit_after_callback(GtkWidget *w, gpointer data);
void modem_attr_callback(GtkWidget *w, gpointer data);
void screen_save_callback(GtkWidget *w, gpointer data);
void chat_callback(GtkWidget *w, gpointer data);
void cancel_callback(GtkWidget *w, gpointer data);
void upload_callback(GtkWidget *w, gpointer data);
void etc_menu_create(GtkWidget *w);
void script_menu_create(GtkWidget *box);
void linefeed_callback(void);
void nohang_exit_callback(void);
void screen_reset(void);
void run_scriptfile(GtkWidget *w, char *filename);
void toggle_local_echo(void);
void toggle_auto_response(GtkWidget *w, gpointer index);
void toggle_remote_echo(void);
void give_shell(void);
void toggle_strip_linefeed(void);
static void write_callback(GtkWidget *w, gpointer data);
static gint write_fragment(gpointer data);
static GtkWidget * control_button_create(GtkWidget *box, Command *command);
static void write_signal_child_handler(int signum);
static void draw_check_indicator(GtkCheckMenuItem *menu, GdkRectangle *area);
static void script_command_ok(GtkWidget *w, GtkWidget *dialog);
static void script_command_cancel(GtkWidget *w, GtkWidget *dialog);
static void query_script_command(void);

static Command command[] = {
	{"", exit_callback, NULL},
#if 0
	{"𵩼", modem_attr_callback, NULL},
#endif
	{"", capture_callback, NULL},
	{"Ϻ", upload_callback, NULL},
	{"ȭ", screen_save_callback, NULL},
	{"", cancel_callback, NULL},
	{"̾߱⸶", chat_callback, NULL},
	{"Ư", si_callback, NULL},
	{"۾", write_callback, NULL},
	{NULL}
};

void
control_bar(GtkWidget *box)
{
	GtkWidget *button;
	gtk_widget_set_name(box, "control");
	gtk_container_border_width(GTK_CONTAINER(box), 3);
	gtk_widget_realize(box);
	button = gtk_button_new();
	GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
	gtk_widget_show(button);
	taegeuk = ani_pixmap_create(button);
	gtk_container_add(GTK_CONTAINER(button), taegeuk);
	gtk_widget_show(taegeuk);
	quit_button = control_button_create(box, command);
	if(use_modem) {
		if(modem_connected()) {
			call_button = gtk_button_new_with_label("ȭ");
		} else {
			call_button = gtk_button_new_with_label("ȭɱ");
		}
	} else {
		call_button = gtk_button_new_with_label("ڳ ");
	}
	gtk_box_pack_start(GTK_BOX(box), call_button, FALSE, FALSE, 0);
	GTK_WIDGET_UNSET_FLAGS(call_button, GTK_CAN_FOCUS);
	gtk_signal_connect(GTK_OBJECT(call_button), "clicked",
		GTK_SIGNAL_FUNC(dial_callback), NULL);
	gtk_widget_show(call_button);
	capture_button = control_button_create(box, command + 1);
	hangup_after_transfer_button = 
		gtk_toggle_button_new_with_label(" ");
	GTK_WIDGET_UNSET_FLAGS(hangup_after_transfer_button, GTK_CAN_FOCUS);
	gtk_box_pack_start(GTK_BOX(box), hangup_after_transfer_button,
		FALSE, FALSE, 0);
	gtk_signal_connect(GTK_OBJECT(hangup_after_transfer_button), "toggled",
		GTK_SIGNAL_FUNC(exit_after_callback), NULL);
	gtk_widget_show(hangup_after_transfer_button);
	gtk_toggle_button_set_state (
		GTK_TOGGLE_BUTTON(hangup_after_transfer_button), FALSE);

#ifdef UPLOAD_BUTTON
	upload_button = control_button_create(box, command + 2);
#endif
	cancel_button = control_button_create(box, command + 4);
	chat_button = control_button_create(box, command + 5);
	control_button_create(box, command + 6);
	write_button = control_button_create(box, command + 7);
	script_menu_create(box);
	etc_menu_create(box);

	if((!use_modem) || !modem_connected()) {
#ifdef UPLOAD_BUTTON
		gtk_widget_set_sensitive(upload_button, FALSE);
#endif
		gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
		gtk_widget_set_sensitive(write_button, FALSE);
	}
}

GtkWidget *
control_button_create(GtkWidget *box, Command *command)
{
	GtkWidget *button = gtk_button_new_with_label(command->label);
	GTK_WIDGET_UNSET_FLAGS(button, GTK_CAN_FOCUS); 
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(command->func), command->data);
	gtk_box_pack_start(GTK_BOX(box), button, FALSE, FALSE, 0);
	gtk_widget_show(button);
	return button;
}

void
capture_callback(GtkWidget *w, gpointer data)
{
	if(capture_fd > 0){
		end_capture();
	} else {
		begin_capture(capture_file);
	}
}


void
hangup_callback(GtkWidget *box, gpointer data)
{
	hangup();
}

void
exit_after_callback(GtkWidget *w, gpointer data)
{
	exit_after_download = GTK_TOGGLE_BUTTON(w)->active;
}
	

void
exit_callback(GtkWidget *w, gpointer data)
{
	if(gau_state == DIAL) {
		dial_cancel();
	} else if(use_modem && modem_connected()) {
		modem_hangup(); modem_check_connected(NULL);
	} else if(!use_modem && telnet_connected()) {
		telnet_hangup(); telnet_time_report(NULL);
	}
	if(use_modem) modem_restore();
	gtk_main_quit();
}

void
modem_attr_callback(GtkWidget *w, gpointer data)
{
	// FIXME!!
}

void
cancel_callback(GtkWidget *w, gpointer data)
{
	switch(gau_state) {
		case SCRIPT:
			script_cancel();
			show_message("ȥڸ ҵǾϴ");
			break;
		case DIAL:
			dial_cancel();
			show_message("ȭɱⰡ ҵǾϴ");
			break;
		case TELNET_DIAL:
			show_message("ڳ  մϴ");
			kill(telnet_pid, SIGTERM);
			break;
		case DOWNLOAD:
			rz_cancel();
			show_message("ޱⰡ  ҵǾϴ");
			break;
		case UPLOAD:
			sz_cancel();
			show_message("øⰡ  ҵǾϴ");
			break;
		default:
			show_message("  ־µ");
			break;
	}
}

void
screen_save_callback(GtkWidget *w, gpointer data)
{
	time_t timep = time(NULL);
	struct tm *lt = localtime(&timep);
	sprintf(tmpbuf, "%s/%02d%02d%02d.scap",expand_path(capture_path),
		lt->tm_year, lt->tm_mon + 1, lt->tm_mday); 
	scapture_fd = open(tmpbuf, O_WRONLY | O_CREAT | O_APPEND, 0660);
	if(scapture_fd == -1) {
		g_warning("can't open screen capture file %s\n", tmpbuf);
		return;
	}
	screen_save();
}

void
upload_callback(GtkWidget *w, gpointer data)
{
	run_sz();
}

void
write_callback(GtkWidget *w, gpointer data)
{
	sprintf(tmpbuf, "%s /tmp/gau%d.txt", editor, getpid());
	editor_pid = fork();
	if(editor_pid == 0) {
		// child
		execlp("sh", "sh", "-c", tmpbuf, NULL);
		g_warning("fail to run %s\n", editor);
		_exit(1);
	}
	signal(SIGCHLD, write_signal_child_handler);
}

void
write_signal_child_handler(int signum)
{
	int fd;
	if(wait(NULL) != editor_pid) {
		signal(SIGCHLD, write_signal_child_handler);
	}
	signal(SIGCHLD, SIG_IGN);
	sprintf(tmpbuf, "/tmp/gau%d.txt", getpid());
	fd = open(tmpbuf, O_RDONLY);
	if(fd < 0) {
		g_warning("fail to read %s\n", tmpbuf);
		return;
	}
	gtk_idle_add(write_fragment, (gpointer)fd);
}

gint
write_fragment(gpointer data)
{
	int fd = (int)data;
	int size, i;
	size = read(fd, tmpbuf + 16, 16);
	if(size > 0) {
		i = 0;
		while(i < size) {
			if(tmpbuf[i + 16] != '\n') tmpbuf[i] = tmpbuf[i+16];
			else tmpbuf[i] = '\r';
			i++;
		}
		term->from_term(tmpbuf, size);
		return TRUE;
	}
	close(fd);
	show_message("ۺ⸦ ϴ");
	sprintf(tmpbuf, "/tmp/gau%d.txt", getpid());
	unlink(tmpbuf);
	return FALSE;
}


#include <glob.h>
void
script_menu_create(GtkWidget *box)
{
	GtkWidget *menubar, *menu, *menuitem;
	GtkWidget *scriptitem;
	char *filename;
	glob_t globd;
	int i;

	menubar = my_menu_bar_new();
	gtk_box_pack_start(GTK_BOX(box), menubar, FALSE, TRUE, 0);
	gtk_container_border_width(GTK_CONTAINER(menubar), 0);
	gtk_widget_show(menubar);

	menu = gtk_menu_new();

	//gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem), menu);

	scriptitem = gtk_menu_item_new_with_label("ɾ");
	gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
		GTK_SIGNAL_FUNC(query_script_command), NULL);
	gtk_menu_append(GTK_MENU(menu), scriptitem);
	gtk_widget_show(scriptitem);

	sprintf(tmpbuf, "%s/*.scr", expand_path(script_path));
	if(0 == glob(tmpbuf, 0, NULL, &globd)) {
		char *basename(char *);
		for(i=0;i<globd.gl_pathc;i++) {
			filename = g_strdup(basename(globd.gl_pathv[i]));
			scriptitem = gtk_menu_item_new_with_label(filename);
			gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
				GTK_SIGNAL_FUNC(run_scriptfile), filename);
			gtk_menu_append(GTK_MENU(menu), scriptitem);
			gtk_widget_show(scriptitem);
		}
		globfree(&globd);
	}

	menuitem = gtk_menu_item_new_with_label("ȥڸ");
	gtk_container_border_width(GTK_CONTAINER(menuitem), 0);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
	gtk_menu_bar_append(GTK_MENU_BAR(menubar), menuitem);
	gtk_widget_show(menuitem);
}


void
etc_menu_create(GtkWidget *box)
{
	GtkWidget *menubar, *menu, *subitem, *menuitem;
	GtkWidget *scriptitem, *scriptitems;

	menubar = my_menu_bar_new();
	gtk_box_pack_start(GTK_BOX(box), menubar, FALSE, TRUE, 0);
	gtk_container_border_width(GTK_CONTAINER(menubar), 0);
	gtk_widget_show(menubar);


	menu = gtk_menu_new();

	if(n_inout) {
		int i;
		subitem = gtk_menu_item_new_with_label("ڵ");
		gtk_menu_append (GTK_MENU(menu), subitem);
		gtk_widget_show(subitem);

		scriptitems = gtk_menu_new();
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem), scriptitems);

		for(i=0;i<n_inout;i++) {
			char *s = g_strdup(inouts[i].raw);
			int max_len = 25;
			if(strlen(s) > max_len) {
				if(is_first_han(s, max_len - 2)) {
					s[max_len - 3] = '.';
				}
				s[max_len - 2] = '.';
				s[max_len - 1] = '.';
				s[max_len] = 0;
			}
			scriptitem = gtk_check_menu_item_new_with_label(s);
			g_free(s);
			gtk_check_menu_item_set_state (GTK_CHECK_MENU_ITEM(scriptitem), inouts[i].active);
			gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
				GTK_SIGNAL_FUNC(toggle_auto_response), (gpointer)i);
			gtk_menu_append(GTK_MENU(scriptitems), scriptitem);
			gtk_widget_show(scriptitem);
		}
	}

	subitem = gtk_check_menu_item_new_with_label("ǵ");
	GTK_CHECK_MENU_ITEM_CLASS(GTK_OBJECT(subitem)->klass)->draw_indicator = draw_check_indicator;
	gtk_signal_connect(GTK_OBJECT(subitem), "activate",
		GTK_SIGNAL_FUNC(linefeed_callback), NULL);
	gtk_menu_append (GTK_MENU(menu), subitem);
	gtk_widget_show(subitem);

	subitem = gtk_menu_item_new_with_label("ȭ ");
	gtk_signal_connect(GTK_OBJECT(subitem), "activate",
		GTK_SIGNAL_FUNC(screen_reset), NULL);
	gtk_menu_append (GTK_MENU(menu), subitem);
	gtk_widget_show(subitem);

	subitem = gtk_menu_item_new_with_label("ȭ ");
	gtk_signal_connect(GTK_OBJECT(subitem), "activate",
		GTK_SIGNAL_FUNC(screen_save_callback), NULL);
	gtk_menu_append (GTK_MENU(menu), subitem);
	gtk_widget_show(subitem);

	subitem = gtk_check_menu_item_new_with_label("^M ");
	gtk_check_menu_item_set_state (GTK_CHECK_MENU_ITEM(subitem), strip_linefeed);
	gtk_signal_connect(GTK_OBJECT(subitem), "activate",
		GTK_SIGNAL_FUNC(toggle_strip_linefeed), NULL);
	gtk_menu_append (GTK_MENU(menu), subitem);
	gtk_widget_show(subitem);

	if(use_modem) {
		subitem = gtk_menu_item_new_with_label("1:1 ");
		gtk_menu_append (GTK_MENU(menu), subitem);
		gtk_widget_show(subitem);

		scriptitems = gtk_menu_new();
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(subitem), scriptitems);

		// FIXME
		scriptitem = gtk_check_menu_item_new_with_label("ڱ");
		gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
			GTK_SIGNAL_FUNC(toggle_local_echo), NULL);
		gtk_menu_append(GTK_MENU(scriptitems), scriptitem);
		gtk_widget_show(scriptitem);
		scriptitem = gtk_check_menu_item_new_with_label("remote echo");
		gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
			GTK_SIGNAL_FUNC(toggle_remote_echo), NULL);
		gtk_menu_append(GTK_MENU(scriptitems), scriptitem);
		gtk_widget_show(scriptitem);

		scriptitem = gtk_menu_item_new_with_label("Ϻ");
		gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
			GTK_SIGNAL_FUNC(upload_callback), NULL);
		gtk_menu_append(GTK_MENU(scriptitems), scriptitem);
		gtk_widget_show(scriptitem);

		scriptitem = gtk_menu_item_new_with_label("Shell");
		gtk_signal_connect(GTK_OBJECT(scriptitem), "activate",
			GTK_SIGNAL_FUNC(give_shell), NULL);
		gtk_menu_append(GTK_MENU(scriptitems), scriptitem);
		gtk_widget_show(scriptitem);
	}

	if(use_modem) {
		subitem = gtk_menu_item_new_with_label("׳ ");
		gtk_signal_connect(GTK_OBJECT(subitem), "activate",
			GTK_SIGNAL_FUNC(nohang_exit_callback), NULL);
		gtk_menu_append (GTK_MENU(menu), subitem);
		gtk_widget_show(subitem);
	}
	menuitem = gtk_menu_item_new_with_label("⵿");
	gtk_container_border_width(GTK_CONTAINER(menuitem), 0);
	gtk_menu_item_set_submenu(GTK_MENU_ITEM(menuitem), menu);
	gtk_menu_bar_append(GTK_MENU_BAR(menubar), menuitem);
	gtk_widget_show(menuitem);
}


void
linefeed_callback(void)
{
	term->auto_linefeed = !term->auto_linefeed;
}

void
toggle_local_echo(void)
{
	local_echo = !local_echo;
}

void
toggle_remote_echo(void)
{
	remote_echo = !remote_echo;
}

void
nohang_exit_callback(void)
{
	extern int modem_sethup(int flag);
	modem_sethup(FALSE);
	_exit(0);
}

void
screen_reset(void)
{
	ansi_term_reset(term);
}

void
run_scriptfile(GtkWidget *w, char *filename)
{
	g_return_if_fail(filename != NULL);
	g_return_if_fail(filename[0] != 0);
	if(filename[0] == '/') {
		strcpy(tmpbuf, filename);
	} else {
		sprintf(tmpbuf, "%s/%s", expand_path(script_path), filename);
	}
	run_script_from_file(tmpbuf);
}

static void
script_command_ok(GtkWidget *w, GtkWidget *dialog)
{
	GtkEntry *entry;
	char *command;
	entry = gtk_object_get_user_data(GTK_OBJECT(w));
	command = gtk_entry_get_text(entry);
	run_script_from_command(command);
	gtk_widget_destroy(dialog);
}

static void
script_command_cancel(GtkWidget *w, GtkWidget *dialog)
{
	gtk_widget_destroy(dialog);
}

static int
script_entry_key_press(GtkEntry *entry, GdkEventKey *ev, GtkWidget *dialog)
{
	char *command;
	if(ev->keyval != GDK_Return) return FALSE;
	command = gtk_entry_get_text(entry);
	run_script_from_command(command);
	gtk_widget_destroy(dialog);
	return TRUE;
}

void
query_script_command(void)
{
	GtkWidget *dialog, *vbox, *hbox, *hse, *button, *label, *entry;

	dialog = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_window_position(GTK_WINDOW(dialog), GTK_WIN_POS_CENTER);
	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(dialog), vbox);
	gtk_widget_show(vbox);
	hbox = gtk_hbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_widget_show(hbox);
	label = gtk_label_new("ɾ:");
	gtk_container_add(GTK_CONTAINER(hbox), label);
	gtk_widget_show(label);

	entry = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(entry), "key_press_event",
		GTK_SIGNAL_FUNC(script_entry_key_press), dialog);
	gtk_container_add(GTK_CONTAINER(hbox), entry);
	gtk_widget_show(entry);

	hse = gtk_hseparator_new();
	gtk_container_add(GTK_CONTAINER(vbox), hse);
	gtk_widget_show(hse);

	hbox = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("");
	gtk_object_set_user_data(GTK_OBJECT(button), entry);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(script_command_ok), dialog);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("");
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(script_command_cancel), dialog);
	gtk_widget_show(button);
	gtk_widget_show(dialog);
}


void
give_shell(void)
{
	// FIXME
	return;
	shell_pid = fork();
	if(shell_pid == 0) {
		// child
		dup2(read_fd, 0);
		dup2(read_fd, 1);
		dup2(read_fd, 2);
		close(read_fd);
		execlp("sh", "sh", "-c", sh_command, NULL);
		g_warning("fail to run %s\n", sh_command);
		_exit(1);
	}
}

void
toggle_strip_linefeed(void)
{
	strip_linefeed = !strip_linefeed;
}

void
toggle_auto_response(GtkWidget *w, gpointer index)
{
	int i = (int)index;
	inouts[i].active = !inouts[i].active;
}

static char *check_d[] = {
/* width height num_colors chars_per_pixel */
"    16    15        4            1",
/* colors */
". c #02027c",
"# c #02fe04",
"a c #027a04",
"b c None",
/* pixels */
"bbbbbbbbbbbbbbbb",
"bbbbbbbbbbbbbbb#",
"bbbbbbbbbbbbbb#a",
"bbbbbbbbbbbbb#aa",
"bbbbbbbbbbbb#aaa",
"bbbbbbbbbbb#aaa.",
"bbbbbbbbbb#aaa.b",
"bbb#bbbbb#aaa.bb",
"bb#a#bbb#aaa.bbb",
"bbaaa#b#aaa.bbbb",
"bb.aaa#aaaa.bbbb",
"bbb.aaaaaa.bbbbb",
"bbbb.aaaa.bbbbbb",
"bbbbb.aa.bbbbbbb",
"bbbbbb..bbbbbbbb"
};

static void
draw_check_indicator(GtkCheckMenuItem *menu, GdkRectangle *area)
{
	static GdkPixmap *check_pixmap = NULL; 
	static GdkBitmap *mask;
	GtkWidget *w;
	int width, height;
	int x, y;
	w = GTK_WIDGET(menu);
	if(check_pixmap == NULL) {
		GdkColor *bg = &w->style->bg[GTK_STATE_NORMAL];
		check_pixmap = gdk_pixmap_create_from_xpm_d(NULL, &mask, bg,
		check_d);
	}
	width = 16;
	height = 15;
	x = 0;
	y = (w->allocation.height - height) / 2;
	gdk_window_clear_area(w->window, x, y, width, height);
	if(menu->active || GTK_WIDGET_STATE(menu) == GTK_STATE_PRELIGHT) {
		gdk_gc_set_clip_mask(w->style->black_gc, mask);
		gdk_gc_set_clip_origin(w->style->black_gc, x, y);
		gdk_draw_pixmap(w->window, w->style->black_gc, check_pixmap, 0, 0, x, y, -1, -1);
		gdk_gc_set_clip_mask(w->style->black_gc, NULL);
		gdk_gc_set_clip_origin(w->style->black_gc, 0, 0);
	}
}

