
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <gtk/gtk.h>
#include "gau.h"

void i_sel_show(AnsiTerm *term);
void i_sel_hide(AnsiTerm *term);
void i_sel_send(AnsiTerm *term);
static void translate(AnsiTerm *term, int *x, int *y);

GdkColor i_sel_col;
GdkGC *i_sel_gc;
static int i_sel_focus;
static int i_sel_begin_col;
static int i_sel_end_col;
static int i_sel_row;
static int i_sel_send_begin;
static int i_sel_send_end;
static char *external_command = NULL;

static int p2(AnsiTerm *term, char *buf, int x);
static int p3(AnsiTerm *term, char *buf, int x);
static int p4(AnsiTerm *term, char *buf, int x);
static int p6(AnsiTerm *term, char *buf, int x);
static int http_p(AnsiTerm *term, char *buf, int x);

void
i_select_motion(AnsiTerm *term, int x, int y)
{
	translate(term, &x, &y);
	if(i_sel_focus) {
		if(i_sel_row != y || i_sel_begin_col > x || i_sel_end_col < x) {
			i_sel_hide(term);
		} 
	} else {
		if(i_sel_row == y && i_sel_begin_col <= x && i_sel_end_col >= x) {
			i_sel_show(term);
		}
	}
}

void
i_select_end(AnsiTerm *term, int x, int y)
{
	if(i_sel_focus) {
		play_click();
		i_sel_hide(term);
		if(external_command) {
			if(fork() == 0) {
				execlp("sh", "sh", "-c", external_command, NULL);
				g_warning("fail to run %s\n", external_command);
				_exit(1);
			}
			g_free(external_command);
			external_command = NULL;
		} else
			i_sel_send(term);
	} else if(external_command) {
		g_free(external_command);
		external_command = NULL;
	}
}

void
i_sel_show(AnsiTerm *term)
{
	i_sel_focus = TRUE;
	gdk_gc_set_foreground(i_sel_gc, &i_sel_col);
	gdk_draw_rectangle(GTK_WIDGET(term)->window, i_sel_gc, TRUE,
		i_sel_begin_col * term->font_width,
		i_sel_row * term->font_height,
		(i_sel_end_col - i_sel_begin_col)* term->font_width,
		term->font_height);
	gdk_gc_set_foreground(i_sel_gc, &GTK_WIDGET(term)->style->black);
	gdk_draw_text(GTK_WIDGET(term)->window, i_sel_gc, 
		i_sel_begin_col * term->font_width, 
		i_sel_row * term->font_height + term->font_ascent,
		term->buf + linenumber(i_sel_row)*term->col + i_sel_begin_col, 
		i_sel_end_col - i_sel_begin_col);
}

void
i_sel_hide(AnsiTerm *term)
{
	GdkRectangle area;
	i_sel_focus = FALSE;
	area.x = i_sel_begin_col * term->font_width; 
	area.y = i_sel_row * term->font_height;
	area.width = (i_sel_end_col - i_sel_begin_col)* term->font_width;
	area.height = term->font_height;
	gtk_widget_draw(GTK_WIDGET(term), &area);
}

void
i_sel_send(AnsiTerm *term)
{
	char *s;
	int size;
	char c[2] = "\r";
	static guint32 oldtime =0;
	guint32 newtime;

	newtime = gdk_time_get();
	if(oldtime + 300 > newtime) {
		oldtime = newtime;
		return;
	}
	s = term->buf + linenumber(i_sel_row)*term->col + i_sel_send_begin;
	size = i_sel_send_end - i_sel_send_begin;
	oldtime = newtime;
	term->from_term(s, size);
	term->from_term(c, 1);
}

int
i_select_begin(AnsiTerm *term, int x, int y)
{
	static char *buf = NULL;
	char *s;
	i_sel_focus = TRUE;

	if(!buf) {
		buf = g_new(char , term->col + 3);
		buf[0] = 0;
		buf++;
		buf[term->col] = 0;
	}
	translate(term, &x, &y);

	i_sel_row = y;
	memcpy(buf, term->buf + linenumber(y)*term->col, term->col);
	/* try matching string with several patterns */

	/* 11998 ӵ   apk27    06/13  14  1 [] Unix   ֽǺ*/
	s = eat_space(buf);
	if(isdigit(*s)) {
		while(isdigit(*++s)) ;
		if(isspace(*s)) {
			i_sel_send_end = s - buf;
			while(isdigit(*--s)) ;
			s++;
			i_sel_begin_col = s - buf;
			i_sel_send_begin = i_sel_begin_col;
			s = buf + term->col -1;
			while(isspace(*s)) s--;
			i_sel_end_col = s - buf + 1;
			i_sel_show(term);
			return TRUE;
		}
	}

	if(p2(term, buf, x)) return TRUE;

	/* ȣ/(H,F,B,P,T,M,N,A,W,DD,AP,UD,PR,DN,S,GO,PF,ME,MM,SAY,Z,X)  
	ȣ/(H. P.ȭ GO.̵ J.ش缭 NO.ü) */

	if(strstr(buf, "ȣ/")) {
		//g_print("2:%s\n", __FUNCTION__);
		s = buf + x;
		while(buf <= s && *s != '.' && *s != ',' && !isspace(*s)) s--;
		if(isspace(*s)) {
			while(*s != '.' && *s != ',') s++;
		} else if(*s == ',') {
			if((buf + x) != s) {
				s++;
				while(*s != ',' && *s != ')') s++;
			}
		}
		i_sel_send_end = s - buf;
		s--;
		while(isalpha(*s)) s--;
		s++;
		i_sel_send_begin = s - buf;
		i_sel_begin_col = s - buf;
		s = buf + i_sel_send_end;
		i_sel_end_col = i_sel_send_end;
		if(*s == '.') {
			s = strchr(s + 1, '.');
			if(s != NULL) {
				s--;
				while(isalpha(*s)) s--;
				while(isspace(*s)) s--;
				i_sel_end_col = s - buf + 1;
			} else {
				s = buf + i_sel_send_end;
				if((s = strchr(s, ')')) != NULL) {
					i_sel_end_col = s - buf;
				}
			}
		}

		i_sel_show(term);
		return TRUE;
	}
	if(p3(term, buf, x)) return TRUE;
	if(p4(term, buf, x)) return TRUE;

	// go linux  , go linux12 
	s = buf + x;
	if(isalnum(*s)) while(isalnum(*s)) s--;
	s++;
	if((s[0] == 'g' || s[0] == 'G') && (s[1] == 'o' || s[1] == 'O')) {
		i_sel_send_begin = s - buf;
		i_sel_begin_col = s - buf;
		s += 3;
		while(isalpha(*s) || isdigit(*s)) s++;
		i_sel_send_end = s - buf;
		i_sel_end_col = s - buf;
		i_sel_show(term);
		return TRUE;
	} else {
		s--;
		while(isspace(*s)) s--;
		if((s[-1] == 'g' || s[-1] == 'G') && (s[0] == 'o' || s[0] == 'O')) {
			s--;
			i_sel_send_begin = s - buf;
			i_sel_begin_col = s - buf;
			s += 3;
			while(isalnum(*s)) s++;
			i_sel_send_end = s - buf;
			i_sel_end_col = s - buf;
			i_sel_show(term);
			return TRUE;
		}
	}
	
	if(!use_modem && http_p(term, buf, x)) return TRUE;

	if(p6(term, buf, x)) return TRUE;
	// fail to match any valid pattern
	// just send 'return'.
	// it will do the good job in most cases

	s = buf;
	i_sel_send_begin = i_sel_send_end = i_sel_begin_col = i_sel_end_col
	= s - buf;
	return FALSE;
}

char *
backstrchr(char *s, char c)
{
	while(*s && *s != c) s--;
	if(*s) return s;
	return NULL;
}

/*          1. HiTEL INFOSHOP                 10. HiNET-P */
//              F1. ȣƮ ڷ               F2. ڷ Խ

int
p2(AnsiTerm *term, char *buf, int x)
{
	char *s;

	s = buf + x;
	while(s != buf && *s != '.')  s--;
	if(s == buf || !isdigit(s[-1])) return FALSE;
	while(isdigit(*--s)) ; 
	if(*s == '.' || *s == '/') return FALSE;
	if(!isspace(s[0]) && ((s[0] & 0x80) == 0)) {
		if(s[0] == '%' || s[0] == '@' || s[0] == '#' || s[0] == '*') {
			i_sel_send_begin = s - buf + 1;
		} else {
			i_sel_send_begin = s - buf;
		}
		s--;
	} else {
		i_sel_send_begin = s - buf + 1;
	}
	s++;
	i_sel_begin_col = s - buf;

	while(isdigit(*++s)) ;
	i_sel_send_end = s - buf;
	while(1) {
		while((!isdigit(*s)) && (*s != 0) && (*s != (char )0xa6)) {
			if(*s & 0x80) s++;
			s++;
		}
		if(*s == 0 || *s == (char)0xa6) {
			s--;
			break;
		}
		while(isdigit(*s)) s++;
		if(*s == '.') {
			s--;
			while(isdigit(*s)) s--;
			if(!isspace(*s) && ((*s & 0x80) == 0)) s--;
			// == '*' || *s == '@' || *s == '#') s--;
			break;
		}
	}
	while(isspace(*s)) s--;
	s++;
	i_sel_end_col = s - buf;
	i_sel_show(term);
	return TRUE;
}

//(P.׸ F. B. PR. RE. Z.) >>  seen in bluecom
static int
p3(AnsiTerm *term, char *buf, int x)
{
	char *s;
	//g_print("%s\n", __FUNCTION__);
	if(strncmp(eat_space(buf), "(", 5) != 0 || x < 6) return FALSE;
	s = buf + x;
	if(strchr(s, ')') == NULL) return FALSE;
	if(isspace(*s)) return FALSE;
	while(!isspace(*s) && *s != '(') s--;
	s++;
	i_sel_begin_col = s - buf;
	i_sel_send_begin = s - buf;
	s = strchr(s, '.');
	if(s == NULL) return FALSE;
	i_sel_send_end = s - buf;
	while(!isspace(*s)) s++;
	i_sel_end_col = s - buf;
	i_sel_show(term);
	return TRUE;
}


		


//ã(FIND) ̵(GO) ʱ޴(T) ޴(M) (X,BYE) Ÿ(Z) 
static int 
p4(AnsiTerm *term, char *buf, int x)
{
	char *e, *b, *s, *m;
	s = buf + x;
	e = strchr(s, ')');
	if(e == NULL) return FALSE;
	b = backstrchr(s, ')');
	if(b == NULL) b = buf;
	else b++;
	b = eat_space(b);
	i_sel_begin_col = b - buf; //
	m = strchr(b, '(');
	if(m == NULL || m > e - 2) return FALSE;
	i_sel_send_begin = m - buf + 1; //
	s = m + 1;
	while(s != e && *s != ')' && *s != ',') s++;
	if((s != e && *s == ',') || (*s == ')' && s == e)) {
		i_sel_send_end = s - buf; //
	} else 
		return FALSE;
	i_sel_end_col = e - buf + 1; //
	i_sel_show(term);
	return TRUE;
}

static int
http_p(AnsiTerm *term, char *buf, int x)
{
	char *b, *e, *s;
	char *h, *f, *t;
    char class[256];
    int i;
	h = f = t = 0;
    for(i=0;i<256;i++) class[i] = 0;
    for(i='a';i<='z';i++) class[i] = 1;
    for(i='A';i<='Z';i++) class[i] = 1;
    for(i='0'; i <= '9'; i++) class[i] = 1;
    class['.']=class['-']=class['/']=class['~']=class['_'] 
	= class[':'] = 1;
	if(external_command) {
		g_free(external_command);
		external_command = NULL;
	}

	s = buf + x;
	while(*s) {
		if(strncmp(s, "http://", 7) == 0) {
			b = s;
			h = s;
			break;
		} else if(strncmp(s, "ftp://", 6) == 0) {
			b = s;
			f = s;
			break;
		} else if(strncmp(s, "telnet://", 9) == 0) {
			b = s;
			t = s;
			break;
		} else
		s--;
	}
	if(*s == 0) {
		s = buf + x + 1;
		while(*s) {
			if(strncmp(s, "http://", 7) == 0) {
				b = s;
				h = s;
				break;
			} else if(strncmp(s, "ftp://", 6) == 0) {
				b = s;
				f = s;
				break;
			} else if(strncmp(s, "telnet://", 9) == 0) {
				b = s;
				t = s;
				break;
			} else
				s++;
		}
	}
	if(*s == 0) return FALSE;

	if(b == h){
		i_sel_begin_col = b - buf;
		e = b + 7;
	    while(class[(unsigned char)(*e)]) e++;
	    *e = 0;
		i_sel_end_col = e - buf;
		sprintf(tmpbuf, "%s %s", http_command, b);
		external_command = g_strdup(tmpbuf);
		i_sel_show(term);
		return TRUE;
	} else if(b == f) {
		i_sel_begin_col = b - buf;
		e = b + 6;
	    while(class[(unsigned char)(*e)]) e++;
	    *e = 0;
		i_sel_end_col = e - buf;
		sprintf(tmpbuf, "%s %s", ftp_command, b);
		external_command = g_strdup(tmpbuf);
		i_sel_show(term);
		return TRUE;
	} else if(b == t) {
		i_sel_begin_col = b - buf;
		e = b + 9;
		while(class[(unsigned char)(*e)]) e++;
		*e = 0;
		i_sel_end_col = e - buf;
		sprintf(tmpbuf, "gaui %s", b + 9);
		external_command = g_strdup(tmpbuf);
		i_sel_show(term);
		return TRUE; 
	}
	return FALSE;
}

//  M1 [//ȸ]      M2 [̹PC]  M3 [/Ʈ]     M4 [õԽù] 
static int 
p6(AnsiTerm *term, char *buf, int x)
{
	char *b;
	char *e;
	if(strstr(buf, "M1") == NULL || strstr(buf, " M2") == NULL) return FALSE;
	b = backstrchr(buf + x, 'M');
	if(b == NULL) {
		b = strchr(buf, 'M');
	}
	i_sel_begin_col = b - buf;
	i_sel_send_begin = b - buf;
	i_sel_send_end = b - buf + 2;
	e = strchr(b, ']');
	if(e == NULL) return FALSE;
	i_sel_end_col = e - buf;
	i_sel_show(term);
	return TRUE;
}

		


void
translate(AnsiTerm *term, int *x, int *y)
{
	if(*x < 0) *x = 0;
	*x = *x / term->font_width;
	if(*x >= term->col) *x = term->col -1;;
	if(*y < 0) *y = 0;
	*y = *y / term->font_height;
	if(*y >= term->row) *y = term->row -1;
}
