
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <time.h>
#include <fcntl.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include "anipixmap.h"
#include "gau.h"
#include "config.h"
enum {SIMPLE_STRING, SIMPLE_SCRIPT, COMPLEX_STRING, COMPLEX_SCRIPT};
int sound_flag = TRUE;
int use_beep = FALSE;
char *gau_path = NULL;
char *sound_path = NULL;
int capture_fd = -1;
int scapture_fd = -1;
int strip_linefeed = FALSE;
int idle_guard_interval = 120 * 1000;
char *play_command = NULL;
int use_ztelnet = 0;
int use_under = FALSE;
int use_bold = FALSE;
char *sz_command = NULL;
char *telnet_command = NULL;
char *ftp_command = NULL;
char *http_command = NULL;
char *capture_path = NULL;/* will be initialized at gau_config */
char *script_path = NULL;
char *capture_file = NULL;
guint32 last_incoming_time = -1;

Auto *inouts;
int n_inout;

int output_ar(char *src, int size, GList *iinfo, GList *rinfo, int type);

static void capture_file_select(void);
static void capture_file_selection_ok(GtkWidget *w, GtkWidget *win);
static void capture_file_selection_cancel(GtkWidget *w, GtkWidget *win);

typedef struct { int type; char *string; int len;} AutoResponse;
int auto_response_complex_filter(char *buf, int byte);


char *
expand_path(char *str)
{
	static char buf[1024];
	static char *home = NULL;
	if(!home) home = getenv("HOME");
	if(str[0] == '~') {
		strcpy(buf, home);
		strcat(buf, str + 1);
		return buf;
	} else {
		return str;
	}
}

char *
get_line(FILE *fp)
{
	int len, tlen;
	static char *buf = NULL;
	static size = 0;
	int conti = FALSE;

	if(size == 0) {
		size = 125;
		buf = g_new(char, size);
	}
	tlen = 0;
	while(fgets(buf + tlen, size - tlen - 1, fp) != NULL) {
		if(buf[tlen] == '#') continue;
		if(conti) {
			char *s;
			s = buf + tlen;
			if(isspace(*s)) {
				while(isspace(*s)) s++;
				memmove(buf + tlen, s, strlen(s) + 1);
			}
		}
		if(buf[tlen] == '\n' || buf[tlen] == 0) {
			if(conti) break;
			else {
				continue;
			}
		}
		conti = FALSE;
		len  =  strlen(buf + tlen);
		if(len == (size - tlen - 2)) { 
			// incomplete read. we need to increase buf size.
			// or eof
			tlen += len;
			size += 125;
			buf = g_realloc(buf, size);
		} else {
			tlen += len;
			if(buf[tlen - 1] == '\n') {
				tlen--;
				buf[tlen] = 0;
			} else {
				if(tlen == (size - 1)) {
					size += 125;
					buf = g_realloc(buf, size);
				}
				continue;
			}
			if(buf[tlen - 1] == '\\' && buf[tlen - 2] != '\\') {
				conti = TRUE;
				tlen--;
				buf[tlen] = 0;
				continue;
			} else {
				break;
			}
		}
	}
	if(tlen == 0) return NULL;
	else return buf;
}


void
play_connect(void)
{
	play("connect.wav");
}

void
play_transfer_end(void)
{
	play("transfer.wav");
}

void
play_exit(void)
{
	play("end.wav");
}

void
play_click(void)
{
	if(use_beep) return;
	play("click.wav");
}

void
play_disconnect(void)
{
	play("disconec.wav");
}

void
play(char *filename)
{
	int pid;
	char buf[256];
	if(use_beep) {
		gdk_beep();
		gdk_beep();
		return;
	}
	if(sound_flag == FALSE) return;
	sprintf(buf,"%s %s/%s ", play_command, expand_path(sound_path), filename);
	pid = fork();
	if(pid == 0) {
		int fd = open("/dev/null", O_WRONLY);
		dup2(fd, 1);
		// child
		execlp("sh", "sh", "-c", buf, NULL);
		g_warning("fail to play %s\n", buf);
		_exit(1);
	} else if(pid < 0) {
		g_warning("fail to fork in %s\n", __FUNCTION__);
	}
}

GtkWidget *
ani_pixmap_create(GtkWidget *w)
{
	GdkPixmap *pixmap;
	GdkBitmap *mask;
	AniItem *item;
	GtkStyle *style = gtk_widget_get_style(w);
	GList *list = NULL;
	int i;
	char filename[60];
	for(i=0;i<6;i++) {
		sprintf(filename, "%s/xpm/flag.xpm.%d", expand_path(gau_path), i);
		pixmap = gdk_pixmap_create_from_xpm(NULL, &mask, 
			&style->bg[GTK_STATE_NORMAL], filename);
		if(pixmap == NULL) {
			g_warning("xpm file open error: %s\n", filename);
			continue;
		}
		item = g_new(AniItem, 1);
		item->pixmap = pixmap;
		item->mask = mask;
		item->duration = 500;
		list = g_list_append(list, item);
	}
	return ani_pixmap_new(list);
}

void
taegeuk_start(void)
{
	ani_start(ANI_PIXMAP(taegeuk));
}

void
taegeuk_stop(void)
{
	ani_stop(ANI_PIXMAP(taegeuk));
}

char *
eat_space(char *buf)
{
	while(isspace(*buf)) buf++;
	return buf;
}

char *
get_string(char *buf)
{
	char *s;
	if(!isspace(*buf)) {
		while(!isspace(*buf)) buf++;
	}
	buf = eat_space(buf);
	if(buf[0] == '"') {
		s = strchr(buf + 1, '"');
		if(s != NULL) *s = 0;
		return g_strdup(buf + 1);
	} 
	s = strchr(buf, ')');
	if(s != NULL) *s = 0;
	return g_strdup(buf);
}

int
get_boolean(char *buf)
{
	if(strstr(buf, "tr") != NULL) return TRUE;
	if(strstr(buf, "fa") != NULL) return FALSE;
	return -1;
}

GdkColor *
color_new(float red, float green, float blue)
{
	GdkColor *color;
	color = g_new(GdkColor, 1);
	color->red = red * 65535.0;
	color->green = green * 65535.0;
	color->blue = blue * 65535.0;
	return color;
}

GdkColor *
get_color(char *s)
{
	float red, green, blue;
	if(!isspace(*s)) while(!isspace(*++s));
	s = eat_space(s);
	sscanf(s, "%f %f %f", &red, &green, &blue);
	return color_new(red, green, blue);
}

int
get_int(char *s)
{
	int i;
	if(!isspace(*s)) while(!isspace(*++s));
	s = eat_space(s);
	sscanf(s, "%d", &i);
	return i;
}

void
gau_config(void)
{
	FILE *fp ;
	char *s;

	/* gtk resource parsing */
	gtk_rc_parse(expand_path(GTK_RC_FILE));

	/* one more parsing for gau's resource with the same file */
	fp = fopen(expand_path(RC_FILE), "r");
	if(fp == NULL) {
		g_warning("fail to open gau's resource file '%s'\n", 
			expand_path(RC_FILE));
		g_warning("gau will use hardcoded default settings\n");
		goto def;
	} 
	while(fgets(tmpbuf, 1024, fp) != NULL) {
		if(tmpbuf[0] != '(') continue;
		/* This line is gau's option */
		s = eat_space(tmpbuf + 1);
		if(strncmp(s, "device", 6) == 0) {
			if(modem_device) continue; // set from command line
			modem_device = get_string(s);
			//g_print("device: '%s'\n", modem_device);
		} else if(strncmp(s, "baud", 4) == 0) {
			modem_speed = get_string(s);
			//g_print("baud: '%s'\n", modem_speed);
		} else if(strncmp(s, "parity", 6) == 0) {
			modem_parity = get_string(s);
			//g_print("parity: '%s'\n", modem_parity);
		} else if(strncmp(s, "bits", 4) == 0) {
			modem_bits = get_string(s);
			//g_print("bits: '%s'\n", modem_bits);
		} else if(strncmp(s, "use_hwf", 7) == 0) {
			modem_use_hwf = get_boolean(s + 7);
			//g_print("use_hwf: %d\n", modem_use_hwf);
		} else if(strncmp(s, "use_swf", 7) == 0) {
			modem_use_swf = get_boolean(s + 7);
			//g_print("use_swf: %d\n", modem_use_swf);
		} else if(strncmp(s, "script_path", 11) == 0) {
			script_path = get_string(s);
		} else if(strncmp(s, "capture_path", 12) == 0) {
			capture_path = get_string(s);
		} else if(strncmp(s, "capture_file", 12) == 0) {
			capture_file = get_string(s);
			//g_print("capture_path: '%s'\n", capture_path);
		} else if(strncmp(s, "use_internal_zmodem", 19) == 0) {
			use_internal_rz = get_boolean(s + 19);
			//g_print("use_internal_rz: %d\n", use_internal_rz);
		} else if(strncmp(s, "do_lock", 7) == 0) {
			do_lock = get_boolean(s + 7);
		} else if(strncmp(s, "saved_lines", 11) == 0) {
			saved_lines = get_int(s);
			// g_print("saved_lines: %d\n", saved_lines);
		} else if(strncmp(s, "use_sound", 9) == 0) {
			sound_flag = get_boolean(s + 9);
		} else if(strncmp(s, "use_beep", 8) == 0) {
			use_beep = get_boolean(s + 8);
		} else if(strncmp(s, "sound_path", 10) == 0) {
			sound_path = get_string(s);
		} else if(strncmp(s, "column", 6) == 0) {
			col = get_int(s);
			if(col <= 20 || col >= 120) col = 82;
		} else if(strncmp(s, "row", 3) == 0) {
			row = get_int(s);
			if(row <= 5 || row >= 50) row = 26;
		} else if(strncmp(s, "font_height", 11) == 0) {
			terminal_font_height = get_int(s);
		} else if(strncmp(s, "font_ascent", 11) == 0) {
			terminal_font_ascent = get_int(s);
		} else if(strncmp(s, "terminal_font", 13) == 0) {
			terminal_font = get_string(s);
		} else if(strncmp(s, "play_command", 12) == 0) {
			play_command = get_string(s);
		} else if(strncmp(s, "silent", 6) == 0) {
			silent = get_boolean(s + 6);
			if(silent  == -1) silent = FALSE;
		} else if(strncmp(s, "black", 5) == 0) {
			black = get_color(s);
		} else if(strncmp(s, "red", 3) == 0) {
			red = get_color(s);
		} else if(strncmp(s, "blue", 4) == 0) {
			blue = get_color(s);
		} else if(strncmp(s, "green", 5) == 0) {
			green = get_color(s);
		} else if(strncmp(s, "yellow", 6) == 0) {
			yellow = get_color(s);
		} else if(strncmp(s, "cyan", 4) == 0) {
			cyan = get_color(s);
		} else if(strncmp(s, "white", 5) == 0) {
			white = get_color(s);
		} else if(strncmp(s, "magenta", 7) == 0) {
			magenta = get_color(s);
		} else if(strncmp(s, "background", 10) == 0) {
			normal_bg = get_int(s);
		} else if(strncmp(s, "foreground", 10) == 0) {
			normal_fg = get_int(s);
		} else if(strncmp(s, "idle_interval", 13) == 0) {
			idle_guard_interval = get_int(s) * 1000;
		} else if(strncmp(s, "sz_command", 10) == 0) {
			sz_command = get_string(s);
		} else if(strncmp(s, "telnet_command", 14) == 0) {
			telnet_command = get_string(s);
		} else if(strncmp(s, "ftp_command", 11) == 0) {
			ftp_command = get_string(s);
		} else if(strncmp(s, "http_command", 12) == 0) {
			http_command = get_string(s);
		} else if(strncmp(s, "use_bold", 8) == 0) {
			use_bold = get_boolean(s);
		} else if(strncmp(s, "use_underline", 13) == 0) {
			use_under = get_boolean(s);
		} else if(strncmp(s, "initialize_string", 17) == 0) {
			initialize_string = get_string(s);
		} else if(strncmp(s, "strip_linefeed", 14) == 0) {
			strip_linefeed = get_boolean(s);
		} else if(strncmp(s, "chat_append", 11) == 0) {
			chat_append = get_boolean(s);
		} else if(strncmp(s, "chat_prefix1", 12) == 0) {
			chat_prefix1 = get_string(s);
		} else if(strncmp(s, "chat_prefix2", 12) == 0) {
			chat_prefix2 = get_string(s);
		} else if(strncmp(s, "chat_suffix", 11) == 0) {
			chat_suffix = get_string(s);
		} else if(strncmp(s, "editor", 6) == 0) {
			editor = get_string(s);
		} else if(strncmp(s, "sh_command", 10) == 0) {
			sh_command = get_string(s);
		} else if(strncmp(s, "can_block", 9) == 0) {
			can_block = get_boolean(s);
		} else if(strncmp(s, "read_hack", 9) == 0) {
			read_hack = get_boolean(s);
		} else {
			g_warning("parsing error : %s\n", tmpbuf);
		}
	}
	def:

	if(capture_path == NULL) capture_path = CAPTURE_PATH;
	if(terminal_font == NULL) terminal_font = DEFAULT_TERMINAL_FONT;
	if(script_path == NULL) script_path = SCRIPT_PATH;
	if(sound_path == NULL) sound_path = DEFAULT_SOUND_PATH;
	if(gau_path == NULL) gau_path = GAU_PATH;
	if(play_command == NULL) play_command = DEFAULT_PLAY_COMMAND;
	if(sz_command == NULL) sz_command = DEFAULT_SZ_COMMAND;
	if(telnet_command == NULL) telnet_command = DEFAULT_TELNET_COMMAND;
	if(*telnet_command == 'z') use_ztelnet = TRUE;
	else use_ztelnet = FALSE;
	if(ftp_command == NULL) ftp_command = DEFAULT_FTP_COMMAND;
	if(http_command == NULL) http_command = DEFAULT_HTTP_COMMAND;
	if(editor == NULL) editor = DEFAULT_EDITOR;
	if(sh_command == NULL) sh_command = DEFAULT_SH;
	if(sound_flag) use_beep = FALSE;
	// color setup
	if(!black) black = color_new(0, 0, 0);
	if(!red) red = color_new(0.7, 0, 0);
	if(!green) green = color_new(0, 0.7, 0);
	if(!yellow) yellow = color_new(0.5, 0.5, 0);
	if(!blue) blue = color_new(0, 0, 0.7);
	if(!magenta) magenta = color_new(0.5, 0, 0.5);
	if(!cyan) cyan = color_new(0, 0.5, 0.5);
	if(!white) white = color_new(0.83, 0.80, 0.93);
	fclose(fp);
}

// buf_siz     ȭ  Ѵ.

#define BUF_SIZ 80*6
char inputbuf[BUF_SIZ];
void
fast_input_handler_from_external(gpointer data, int source, 
	GdkInputCondition con)
{
	int size, newsize;
	int retval;
	typedef int (*FilterFunc)(char *str, int bytes);
	GList *list = input_filter_list;
	long lf;
	static guint32 last_time = 0;
	guint32 this_time;
	show_incoming(TRUE);
	// Ʒ  250 70  33.6K𵩿 ߾ Ӵϴ.
	// ٸ ӵ  ׿  ִ   ӵ
	//    ϴ.
	this_time = gdk_time_get();
	size = read(source, inputbuf, BUF_SIZ);
	if(this_time - last_time < 250) {
		while(1) {
			usleep(70);
			ioctl(source, FIONREAD, &lf);
			if(lf == 0) break;
			newsize = read(source, inputbuf + size, BUF_SIZ - size);
			//g_print("%d\n", newsize);
			if(newsize <= 0) break;
			size += newsize;
			if(size >= BUF_SIZ) break;
		}
		last_time = gdk_time_get();
	} else {
		last_time = this_time;
	}
	
	while(list) {
		FilterFunc func;
		func = list->data;
		list = list->next;
		retval = (*func)(inputbuf, size);
		if(retval) return; // eaten
	}
	if(remote_echo) {
		int i;
		char c = '\n';
		for(i=0;i<size;i++) {
			if(inputbuf[i] != '\r') {
				term->from_term(inputbuf + i, 1);
			} else {
				term->from_term(inputbuf + i, 1);
				term->from_term(&c, 1);
			}
		}
	}
	term->to_term(term, inputbuf, size);
	show_incoming(FALSE);
	last_incoming_time = gdk_time_get();
}

#undef BUF_SIZ

void
input_handler_from_external(gpointer data, int source, GdkInputCondition con)
{
	int size;
	int retval;
	typedef int (*FilterFunc)(char *str, int bytes);
	GList *list = input_filter_list;
	show_incoming(TRUE);
	size = read(source, inputbuf, 2048);
	
	while(list) {
		FilterFunc func;
		func = list->data;
		list = list->next;
		retval = (*func)(inputbuf, size);
		if(retval) return; // eaten
	}
	if(remote_echo) {
		int i;
		char c = '\n';
		for(i=0;i<size;i++) {
			if(inputbuf[i] != '\r') {
				term->from_term(inputbuf + i, 1);
			} else {
				term->from_term(inputbuf + i, 1);
				term->from_term(&c, 1);
			}
		}
	}
	term->to_term(term, inputbuf, size);
	show_incoming(FALSE);
	last_incoming_time = gdk_time_get();
}

void
hangup(void)
{
	if(use_modem) modem_hangup();
	else telnet_hangup();
}

int
capture_input_filter(char *str, int bytes)
{
	int i;
	if(strip_linefeed) {
		for(i=0;i<bytes;i++) if(str[i] != '\r') write(capture_fd, str + i, 1);
	} else
		write(capture_fd, str, bytes);
	return FALSE;
}

void
begin_capture(char *filename)
{
	if(filename == NULL || filename[0] == 0) {
		time_t timep = time(NULL);
		struct tm *lt = localtime(&timep);
		sprintf(tmpbuf, "%s/%02d%02d%02d.cap",expand_path(capture_path),
			lt->tm_year, lt->tm_mon + 1, lt->tm_mday); 
		capture_fd = open(tmpbuf, O_WRONLY | O_CREAT | O_APPEND, 0660);
		if(capture_fd == -1) {
			g_warning("can't open capture file %s\n", tmpbuf);
			return;
		}
	} else if(strcmp(filename, "query") == 0) {
		capture_file_select();
		return;
	} else {
		if(filename[0] == '/') strcpy(tmpbuf, filename);
		else sprintf(tmpbuf, "%s/%s", expand_path(capture_path), filename);
		capture_fd = open(tmpbuf, O_WRONLY | O_CREAT | O_APPEND, 0660);
		if(capture_fd < 0) {
			g_warning("can't open capture file %s\n", tmpbuf);
			return;
		}
	}
	input_filter_list = g_list_append(input_filter_list, capture_input_filter);
	gtk_label_set(GTK_LABEL(GTK_BUTTON(capture_button)->child), "");
}

static void
capture_file_select(void)
{
	GtkWidget *file_selection;
	file_selection = gtk_file_selection_new(" ");
	sprintf(tmpbuf, "%s/", capture_path);
	gtk_file_selection_set_filename (GTK_FILE_SELECTION(file_selection), tmpbuf);
	gtk_window_position(GTK_WINDOW(file_selection), GTK_WIN_POS_CENTER);
	gtk_signal_connect(
		GTK_OBJECT(GTK_FILE_SELECTION(file_selection)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(capture_file_selection_ok), 
		file_selection);
	gtk_signal_connect(
		GTK_OBJECT(GTK_FILE_SELECTION(file_selection)->cancel_button),
		"clicked",
		GTK_SIGNAL_FUNC(capture_file_selection_cancel), 
		file_selection);
	gtk_widget_destroy(GTK_FILE_SELECTION(file_selection)->help_button);
	gtk_widget_show(file_selection);
}

static void
capture_file_selection_ok(GtkWidget *w, GtkWidget *win)
{
	char *filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(win));
	//g_print("filename: %s\n", filename);
	begin_capture(filename);
	gtk_widget_destroy(win);
}

static void
capture_file_selection_cancel(GtkWidget *w, GtkWidget *win)
{
	gtk_widget_destroy(win);
}


void
end_capture()
{
	input_filter_list = g_list_remove(input_filter_list, capture_input_filter);
	close(capture_fd);
	gtk_label_set(GTK_LABEL(GTK_BUTTON(capture_button)->child), "");
	capture_fd = -1;
}


void
screen_save(void)
{
	char *buf;
	char c = '\n';
	int i;
	int byte;

	for(i=0;i<term->max_reached;i++) {
		buf = term->buf + linenumber(i)*term->col;
		byte = term->col - 1;
		while(byte>=0) {
			if(buf[byte] != ' ') break;
			byte--;
		}
		byte++;
		write(scapture_fd, buf, byte);
		write(scapture_fd, &c, 1);
	}
	close(scapture_fd);
}

int
delayed_real_run(void (*func)(void))
{
	(*func)();
	return FALSE;
}

void
delayed_run(guint32 msec, void (*func)(void))
{
	gtk_timeout_add(msec, (GtkFunction)delayed_real_run, (gpointer)func);
}

inline int
is_substring(register guchar *h, int size, char  *str, int msize, int *jm)
{
	register int j;
	guchar *hend;
	if(size < msize) return 0;
	hend = h + size;
	h += msize - 1;
	do {
		j = jm[*h];
		if(j == 0) {
			switch(j = jm[*(h-1)]) {
				case 1: case 0:
					if(strncmp(h - msize + 1, str, msize) == 0) return 1;
					h++;
					break;
				default:
					h += j - 1;
					break;
			}
		} else {
			h += j;
		}
	} while(h < hend );
	return 0;
}


void
prepare_jm(int msize, char *str, int *jm)
{
	int i;
	for(i=0;i<256;i++) jm[i] = msize;
	for(i=0;i<msize;i++) {
		jm[str[i]] = msize - i - 1;
	}
}

void
terminal_line_changed(AnsiTerm *term)
{
	auto_response_check(term->buf + term->col*linenumber(term->cu_y), term->cu_x);
}

int
auto_response_check(char *str, int size)
{
	int i;
	Auto *inout = inouts;
	int result = FALSE;

	for(i=0;i<n_inout;i++) {
		if(inout->active && is_substring(str, size, inout->lpattern, inout->llen, inout->jm)) {
			switch(inout->type) {
				case COMPLEX_SCRIPT: case COMPLEX_STRING:
					result = output_ar(str, size, 
						inout->ilist, inout->olist, inout->type);
					break;
				case SIMPLE_SCRIPT:
					result = TRUE;
					run_script_from_command((char *)inout->olist);
					break;
				case SIMPLE_STRING:
					result = TRUE;
					term->from_term((char *)inout->olist, strlen((char *)inout->olist));
					break;
			}
			if(result) return TRUE;
		}
		inout++;
	}
	return FALSE;
}


int
pattern_remain(GList *item)
{
	AutoResponse *p;
	if(item->next == NULL) return FALSE;
	item = item->next;
	while(item) {
		p = item->data;
		if(p->type == 0) return TRUE;
		item = item->next;
	}
	return FALSE;
}


int
output_ar(char *src, int size, GList *iinfo, GList *rinfo, int type)
{
	char name[5][20];
	char buf[200];
	char *s;
	char *old_s;
	char *di;
	AutoResponse *pattern;
	int index;
	memcpy(buf, src, size);
	buf[size] = 0;

	index = -1;
	while(iinfo) {
		pattern = iinfo->data;
		if(pattern->type == 0) break;
		else index = pattern->type;
		iinfo = iinfo->next;
	}
	s = old_s = buf;
	while(*s && iinfo) {
		s = strstr(s, pattern->string);
		if(s == NULL) return FALSE;
		if(index > 0) {
			di = name[index - 1];
			memmove(di, old_s, s - old_s);
			di[s - old_s] = 0;
		}
		s += strlen(pattern->string);
		old_s = s;

		iinfo = iinfo->next;
		while(iinfo) {
			pattern = iinfo->data;
			if(pattern->type == 0) break;
			else index = pattern->type;
			iinfo = iinfo->next;
		}
	}
	s = tmpbuf;
	while(rinfo) {
		pattern = rinfo->data;
		if(pattern->type == 0) {
			if(type == COMPLEX_STRING) term->from_term(pattern->string, strlen(pattern->string));
			else s = stpcpy(s, pattern->string);
			printf("%s", pattern->string);
		} else {
			if(type == COMPLEX_STRING) term->from_term(name[pattern->type - 1], strlen(name[pattern->type -1]));
			else s = stpcpy(s, name[pattern->type - 1]);
			printf("%s", name[pattern->type - 1]);
		}
		rinfo = rinfo->next;
	}
	//g_print("script:%s\n", tmpbuf);
	if(type) run_script_from_string(tmpbuf);
	return TRUE;
}

void
auto_response_prepare(void)
{
	char filename[256];
	char p_buf[100] = {0};
	FILE *fp;
	Auto *inout;
	GList *input_list, *output_list;
	AutoResponse *input;
	int n;
	char *s, *buf, *p;
	sprintf(filename, "%s/auto_response", expand_path(gau_path));
	fp = fopen(filename, "r");
	if(fp == NULL) {
		//g_warning("can't open auto_response data file");
		return;
	}
	// check # of pattern
	n_inout = 0;
	while((buf = get_line(fp)) != NULL) {
		n_inout++;
	}
	// check whether file is empty
	if(n_inout == 0) return; 

	// alloc the neccesary memory for pattern;
	inouts = g_new(Auto, n_inout);

	// reread the file
	rewind(fp);
	n = 0;
	inout = inouts;
	while((buf = get_line(fp)) != NULL) {
		if(strncmp(buf, "off", 3) == 0) {
			inout->active = FALSE;
			s = eat_space(buf + 3);
		} else if(strncmp(buf, "on", 2) == 0) {
			inout->active = TRUE;
			s = eat_space(buf + 2);
		} else { 
			inout->active = TRUE;
			s = buf;
		}
		inout->raw = g_strdup(s);
		while(*s) {
			if(*s == '%' && isdigit(*(s + 1))) {
				break;
			}
			s++;
		}
		if(*s == '%') {
			// complex pattern;

			p = p_buf;
			*p = 0;
			input_list = NULL;
			output_list = NULL;

			// extract input pattern information

			s = strchr(buf, '"') + 1;
			//g_print("inputpattern: %s\n", s);
			while(1) {
				if(*s == '%' && isdigit(*(s+1))) {
					if(p_buf[0] != 0) {
						input = g_new(AutoResponse, 1);
						input->type = 0;
						*p = 0;
						transform_string(p_buf, &input->string);
						input_list = 
							g_list_append(input_list, input);
						//g_print("input: %s\n", input->string);
					}
					input = g_new(AutoResponse, 1);
					input->type = *(s+1) - '0';
					s += 2;
					p = p_buf;
					*p = 0;
					input_list = 
						g_list_append(input_list, input);
				} else if(*s == '"') {
					// end of input pattern
					if(p_buf[0] != 0) {
						input = g_new(AutoResponse, 1);
						input->type = 0;
						*p = 0;
						transform_string(p_buf, &input->string);
						input_list = 
							g_list_append(input_list, input);
						//g_print("input: %s\n", input->string);
					}
					s++;
					break;
				} else if(*s == '\\') {
					*p++ = *s++;
					*p++ = *s++;
				} else {
					*p++ = *s++;
				}
			}

			inout->ilist = input_list;

			{
				GList *list = input_list;
				inout->llen = 0;
				//g_print("starti%d: ", nc);
				while(list) {
					input = list->data;
					if(input->type == 0) {
						if(inout->llen < strlen(input->string)) {
							inout->lpattern = input->string;
							inout->llen = strlen(input->string);
						}
						//g_print("%s\n", input->string);
						break;
					}
					list = list->next;
				}
			}


			// extract output pattern information
			s = eat_space(s);
			if(*s == '"') {
				s++;
				inout->type  = COMPLEX_STRING;
			} else {
				inout->type = COMPLEX_SCRIPT;
			}
			//g_print("%d: %d\n", auto_complex_type[nc], nc);
			p = p_buf;
			*p = 0;
			while(1) {
				if(*s == '%' && isdigit(*(s+1))) {
					if(p_buf[0] != 0) {
						input = g_new(AutoResponse, 1);
						input->type = 0;
						*p = 0;
						if(inout->type) {
							input->string = g_strdup(p_buf);
						} else 
							transform_string(p_buf, &input->string);
						output_list = 
							g_list_append(output_list, input);
						//g_print("output: %s\n", input->string);
					}
					input = g_new(AutoResponse, 1);
					input->type = *(s+1) - '0';
					s += 2;
					p = p_buf;
					*p = 0;
					output_list = 
						g_list_append(output_list, input);
				} else if(*s == 0 || (inout->type == 0 && *s == '"')) {
					// end of output pattern
					if(p_buf[0] != 0) {
						input = g_new(AutoResponse, 1);
						input->type = 0;
						*p = 0;
						if(inout->type) {
							input->string = g_strdup(p_buf);
						} else 
							transform_string(p_buf, &input->string);
						output_list = 
							g_list_append(output_list, input);
						//g_print("output: %s\n", input->string);
					}
					break;
				} else if(*s == '\\') {
					*p++ = *s++;
					*p++ = *s++;
				} else {
					*p++ = *s++;
				}
			}
			inout->olist = output_list;
		} else {
			// input pattern parsing
			s = strchr(buf, '"') + 1;
			s = transform_string(s, &inout->lpattern);
			inout->llen = strlen(inout->lpattern);
			prepare_jm(inout->llen, inout->lpattern, inout->jm);
			// output string parsing
			s = eat_space(s + 1);
			if(*s == '"') {
				inout->type = SIMPLE_SCRIPT;
				s++;
				transform_string(s, (char **)&inout->olist);
			} else {
				inout->type = SIMPLE_STRING;
				inout->olist = (GList *)g_strdup(s);
			}
		}
		inout++;
	}
	fclose(fp);
}

char *
transform_string(char *s, char **des)
{
	char *tmp = tmpbuf;
	while(*s && *s != '"') {
		if(*s == '\\') {
			s++;
			if(*s == 'r') {
				*tmp++ = '\r';
				s++;
			} else if(*s == 't') {
				*tmp++ = '\t';
				s++;
			} else if(*s == 'n') {
				*tmp++ = '\n';
				s++;
			} else
				*tmp++ = *s++;
		} else
			*tmp++ = *s++;
	}
	*tmp = 0;
	*des = g_strdup(tmpbuf);
	return s;
}

