#include <stdio.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <ctype.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <gdk/gdkprivate.h>

#include <gdk/automata.h>
#include "word.h"
#define MAX_LEVEL 9
typedef struct { gchar *string;
				 guchar x;
				 guchar y;
				 guchar size;
} WordInfo;

int word_data_parse(void);
void game_init(void);
void choose_level(void);
void green_make(void);
void green_draw(void);
void end_dialog(void);
void inform_new_level(int level);
GtkWidget *drawing;
GtkWidget *entry;
int col, row;
char **word;
int nword;
void show_state(void);
void begin_game(void);
int game_pause(void);
int move_function(gpointer user_data);
int create_function(gpointer user_data);
GList * word_escape(GList *list);
GList * destroy_word_item(GList *list);
GList * destroy_word(GList *list);
int find_match(void);
int key_press(GtkWidget *w, GdkEventKey *ev);
int expose(GtkWidget *w, GdkEventExpose *ev);
void erase(WordInfo *);
void draw(WordInfo *);
extern int check_collision(WordInfo *);
int rand_int(int);
void destroy_callback(GtkWidget *w, gpointer user_data);
extern char *basename(char *);

int width, height;

GtkWidget *state_label1, *state_label2, *state_label3, *state_label4;
GtkWidget *pause_button;
GtkWidget *level_label;
int n_visible_word = 0;
int n_destroyed_word = 0;
int n_failed_word = 0;
int n_appear_word = 0;
int score = 0;
int level = 0;
int n_word_this_level;
int n_remaining_green;
int paused = FALSE;
char state[20];

GdkGC *drawing_fg_gc, *drawing_bg_gc, *drawing_bg_word;
GdkGC *green_gc, *red_gc;
GdkFont *drawing_font;
int font_height, font_width;
gint create_tag, move_tag;
int create_interval[MAX_LEVEL] = {
	6000,
	5000,
	4500,
	4000,
	2500,
	1300,
	900,
	600,
};
int move_interval[MAX_LEVEL] = {
	5500,
	4500,
	4000,
	3200,
	2000,
	1100,
	700,
	400,
};
int n_word[MAX_LEVEL] = {
	2,
	30,
	40,
	50,
	60,
	100,
	150,
	2000
};

gboolean all_flushed;
GList *word_list = NULL;

int
main(int argc, char *argv[])
{
	char rcfile[256];
	if(!word_data_parse()) {
		g_error("word_file");
		return 1;
	}
	sprintf(rcfile, "%s/.httrc", getenv("HOME"));
	gtk_init(&argc, &argv);
	gtk_rc_parse(rcfile);
	col = 30;
	row = 25;
	game_init();
	gtk_main();
	return 0;
}



void
game_init(void)
{
	GtkWidget *win;
	GtkStyle *style;
	GtkWidget *vbox, *hbox, *frame, *separator;
	GdkColor red_color, green_color;
	gint old_mask;
	GdkColormap *colormap =
		gdk_window_get_colormap((GdkWindow *) & gdk_root_parent);

	srand(time(NULL));

	win = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_window_position(GTK_WINDOW(win), GTK_WIN_POS_CENTER);
	gtk_window_set_policy(GTK_WINDOW(win), FALSE, FALSE, FALSE);
	gtk_container_border_width(GTK_CONTAINER(win), 5);
	gtk_signal_connect(GTK_OBJECT(win), "destroy",
		GTK_SIGNAL_FUNC(destroy_callback), NULL);
	gtk_window_set_title(GTK_WINDOW(win), basename(gdk_progname));

	vbox = gtk_vbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(win), vbox);
	gtk_widget_show(vbox);


	hbox = gtk_hbox_new(FALSE, 4);
	gtk_box_pack_end(GTK_BOX(vbox), hbox, FALSE, FALSE, FALSE);
	gtk_widget_show(hbox);
	separator = gtk_hseparator_new();
	gtk_box_pack_end(GTK_BOX(vbox), separator, FALSE, FALSE, FALSE);
	gtk_widget_show(separator);

	pause_button = gtk_button_new_with_label(" ");
	gtk_signal_connect(GTK_OBJECT(pause_button), "clicked",
		GTK_SIGNAL_FUNC(game_pause), NULL);
	gtk_box_pack_start(GTK_BOX(hbox), pause_button, FALSE, FALSE, 0);
	gtk_widget_show(pause_button);


	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	entry = gtk_entry_new();
	gtk_signal_connect(GTK_OBJECT(entry), "key_press_event",
		GTK_SIGNAL_FUNC(key_press), NULL);
	gtk_container_add(GTK_CONTAINER(frame), entry);
	gtk_entry_set_text(GTK_ENTRY(entry), "");
	gtk_widget_show(entry);

	

	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	hbox = gtk_hbox_new(TRUE, 2);
	gtk_container_add(GTK_CONTAINER(frame), hbox);
	gtk_widget_show(hbox);

	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	level_label = gtk_label_new("1 ܰ");
	gtk_widget_set_sensitive(level_label, FALSE);
	gtk_container_add(GTK_CONTAINER(frame), level_label);
	gtk_widget_show(level_label);


#if 0
	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	state_label1 = gtk_label_new("");
	gtk_container_add(GTK_CONTAINER(frame), state_label1);
	gtk_widget_show(state_label1);

	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	state_label2 = gtk_label_new("");
	gtk_container_add(GTK_CONTAINER(frame), state_label2);
	gtk_widget_show(state_label2);

	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	state_label3 = gtk_label_new("");
	gtk_container_add(GTK_CONTAINER(frame), state_label3);
	gtk_widget_show(state_label3);
#endif

	frame = gtk_frame_new(NULL);
	gtk_container_add(GTK_CONTAINER(hbox), frame);
	gtk_widget_show(frame);
	state_label4 = gtk_label_new("");
	gtk_container_add(GTK_CONTAINER(frame), state_label4);
	gtk_widget_set_sensitive(state_label4, FALSE);
	gtk_widget_show(state_label4);

	style = gtk_widget_get_style(entry);
	drawing_font = style->font;
	font_height = drawing_font->ascent + drawing_font->descent;
	font_width = gdk_string_width(drawing_font, "");
	width = font_width * col;
	height = font_height * row;

	drawing = gtk_drawing_area_new();
	
	old_mask = gtk_widget_get_events (drawing);
	gtk_widget_set_events (drawing, old_mask | GDK_EXPOSURE_MASK);
	gtk_drawing_area_size(GTK_DRAWING_AREA(drawing), width, height);
	gtk_box_pack_start(GTK_BOX(vbox), drawing, FALSE, FALSE, FALSE);
	
	/*
	GTK_WIDGET_CLASS(GTK_OBJECT(drawing)->klass)->expose_event = expose;
		*/
	gtk_signal_connect(GTK_OBJECT(drawing), "expose_event",
		GTK_SIGNAL_FUNC(expose), NULL);
	gtk_widget_show(drawing);
	gtk_widget_show(win);

	green_color.red = 0;
	green_color.green = 40000;
	green_color.blue = 0;
	gdk_color_alloc(colormap, &green_color);
	red_color.red = 50000;
	red_color.green = 20000;
	red_color.blue = 20000;
	gdk_color_alloc(colormap, &red_color);

	green_gc = gdk_gc_new(drawing->window);
	gdk_gc_set_foreground(green_gc, &green_color);
	red_gc = gdk_gc_new(drawing->window);
	gdk_gc_set_foreground(red_gc, &red_color);

	show_state();
	choose_level();
}

void
begin_game(void)
{
	char s[10];
	all_flushed = FALSE;
	n_visible_word = 0;
	n_appear_word = 0;
	n_word_this_level = 0;
	sprintf(s, "%d ܰ", level + 1);
	gtk_label_set(GTK_LABEL(level_label), s);
	if(n_failed_word == 0 && n_destroyed_word == 0) {
		int i= -1;
		while(++i <= level) n_word_this_level += n_word[i]; 
	} else 
		n_word_this_level = n_word[level];
	create_function(NULL);
	gtk_timeout_add(move_interval[level], move_function, NULL);
	gtk_timeout_add(create_interval[level], create_function, NULL);
	gtk_widget_set_sensitive(entry, TRUE);
	gtk_widget_grab_focus(entry);
}

int
key_press(GtkWidget *w, GdkEventKey *event)
{
	if(paused && event->keyval < 0xff) game_pause();
	if(event->state != 0) return FALSE;
	if(event->keyval == GDK_space || event->keyval == GDK_Return ||
		event->keyval == GDK_Tab) {
		automata_clear();
		if(find_match()){
			show_state();
		}

		gtk_entry_set_text(GTK_ENTRY(entry), "");
		gtk_signal_emit_stop_by_name(GTK_OBJECT(w), "key_press_event");
		return TRUE;
	}
	return FALSE;
}

int
find_match(void)
{
	char *new_word = gtk_entry_get_text(GTK_ENTRY(entry));
	GList *list = word_list;
	WordInfo *w;
	int matched = FALSE;
	while(list) {
		w = (WordInfo *)list->data;
		if(strcmp(new_word, w->string) == 0) {
			list = destroy_word(list);
			matched = TRUE;
			continue;
		}
		list = list->next;
	}
	return matched;
}

GList *
destroy_word(GList *list)
{
	WordInfo *w = list->data;
	score += (level + 1)*w->size*5;
	erase(w);
	list = destroy_word_item(list);
	n_destroyed_word++;
	n_visible_word--;
	return list;
}

GList *
destroy_word_item(GList *list)
{
	GList * next = list->next;
	word_list = g_list_remove_link(word_list, list);
	g_free(list->data);
	g_list_free(list);
	return next;
}

int
move_function(gpointer user_data)
{
	GList *list = word_list;
	WordInfo *w;
	gboolean collided = FALSE;
	if(n_remaining_green == 0) return FALSE;
	if(paused) return TRUE;
	while(list) {
		w = (WordInfo *)(list->data);
		if(check_collision(w)) {
			/* collide */
			list = word_escape(list);
			collided = TRUE;
			continue;
		} else {
			/* */
			erase(w);
			w->y++;
			draw(w);
		};
		list = list->next;
	}
	/* register next move function */
	if(all_flushed && n_visible_word == 0 ) {
		score += (level + 1) * 100;
		level++;
		if(level == MAX_LEVEL) choose_level();
		else inform_new_level(level);
	} else if(n_remaining_green == 0) {
		gtk_widget_set_sensitive(entry, FALSE);
		end_dialog();
	} else  {
		if(collided) show_state();
		return TRUE;
	}
	return FALSE;
}

GList *
word_escape(GList *list)
{
	WordInfo *w = list->data;
	n_visible_word--;
	n_failed_word++;
	score -= (level + 1) * 3;
	erase(w);
	return destroy_word_item(list);
}

int
create_function(gpointer user_data)
{
	int new_word_index;
	WordInfo *w;
	int i;
	extern char **board;
	if(n_remaining_green == 0) return FALSE;
	if(paused) return TRUE;

	new_word_index = rand_int(nword);
	w = g_new(WordInfo, 1);
	w->string = word[new_word_index];
	w->size = strlen(w->string)/2;
	if(n_remaining_green < 15 && (rand_int(2) == 0)) {
		int j = 0;
		for(i=0;i<col;i++) if(board[i][row-1]) j++;
		j = rand_int(j);
		for(i=0;i<col;i++) {
			if(board[i][row-1]) {
				if(j==0) { 
					w->x =  i;
					break;
				}
				j--;
			}
		}
		if(w->x + w->size > col - 1 ) w->x = col - w->size;
	} else 
		w->x = rand_int(col - w->size + 1);
	w->y = 0;
	
	word_list = g_list_append(word_list, w);
	n_appear_word++;
	n_visible_word++;

	draw(w);
	show_state();

	if((n_appear_word < n_word_this_level) && n_remaining_green ) {
		return TRUE;
	} else if(n_remaining_green == 0) {
	} else  {
		all_flushed = TRUE;
	}
	return FALSE;
}

void
erase(WordInfo *w)
{
	if(!drawing_bg_gc) {
		GtkStyle *style = gtk_widget_get_style(entry);
		drawing_bg_gc = style->bg_gc[GTK_STATE_NORMAL];
	}
	gdk_window_clear_area(drawing->window, 
		w->x * font_width, w->y * font_height, w->size * font_width, 
		font_height);
}

void
draw(WordInfo *w)
{
	if(!drawing_fg_gc) {
		GtkStyle *style = gtk_widget_get_style(drawing);
		GdkColormap *colormap = gdk_window_get_colormap((GdkWindow *) &gdk_root_parent); 
		GdkGCValues values;
		GdkColor fg, bg;
		drawing_fg_gc = gdk_gc_new(drawing->window);
		drawing_bg_word = gdk_gc_new(drawing->window);

		fg.red = 60000;
		fg.green = 40000;
		fg.blue = 40000;

		bg.red = 5000;
		bg.green = 20000;
		bg.blue = 5000;

		gdk_color_alloc(colormap, &fg);
		gdk_color_alloc(colormap, &bg);
		gdk_gc_set_foreground(drawing_fg_gc, &fg);
		gdk_gc_get_values(style->fg_gc[GTK_STATE_NORMAL], &values);
		gdk_gc_set_font(drawing_fg_gc, drawing_font);
		gdk_gc_set_foreground(drawing_bg_word, &bg);
	}
	gdk_draw_rectangle(drawing->window, drawing_bg_word, TRUE,
		w->x * font_width, w->y * font_height, w->size * font_width, font_height);
	gdk_draw_text(drawing->window, drawing_fg_gc, 
		w->x * font_width, w->y * font_height + drawing_font->ascent, 
		w->string, w->size*2);
}


void
destroy_callback(GtkWidget *w, gpointer user_data)
{
	gtk_main_quit();
}

int
rand_int(int n)
{
	return (int)
	((double)n*rand()/((double)RAND_MAX + 1));
}

void
show_state(void)
{
#if 0
	sprintf(state, "%d", n_appear_word);
	gtk_label_set(GTK_LABEL(state_label1), state);
	sprintf(state, "%d", n_failed_word);
	gtk_label_set(GTK_LABEL(state_label2), state);
	sprintf(state, "%d", n_destroyed_word);
	gtk_label_set(GTK_LABEL(state_label3), state);
#endif
	sprintf(state, "%d", score);
	gtk_label_set(GTK_LABEL(state_label4), state);
	
}

gint
expose(GtkWidget *widget, GdkEventExpose *ev)
{
	GdkRectangle word_rec, intersect;
	GList *list = word_list;
	WordInfo *w;
	while(list) {
		w = (WordInfo *)list->data;
		word_rec.x = w->x * font_width;
		word_rec.y = w->y * font_height;
		word_rec.width = w->size * font_width;
		word_rec.height = font_height;
		if(gdk_rectangle_intersect(&ev->area, &word_rec, &intersect)) {
			draw(w);;
		}
		list = list->next;
	}
	green_draw();
	return FALSE;
}

int
game_pause(void)
{
	paused = paused ? FALSE: TRUE;
	if(paused) {
		/*
		gtk_widget_grab_focus(pause_button);
		*/
		gtk_label_set(GTK_LABEL(
			GTK_BUTTON(pause_button)->child), " ϱ");
	} else {
		/*
		gtk_widget_grab_focus(entry);
		*/
		gtk_label_set(GTK_LABEL(
			GTK_BUTTON(pause_button)->child), " ");
	}
	gtk_widget_grab_focus(entry);
	return 0;
}


#define N_GREEN 30
void green_draw(void);
int check_collision(WordInfo *w);
void draw_board(int x, int y);



char **board = NULL;
void
green_make(void)
{
	int i,j;
	int n_green = N_GREEN;
	n_remaining_green = N_GREEN;
	if(board == NULL) {
		board = g_new(char *, col);
		for(i=0;i<col;i++) board[i] = g_new(char, row);
	}
	for(i=0;i<col;i++) for(j=0;j<row;j++) board[i][j] = 0;
	while(n_green--) {
		i = rand_int(col);
		j = row - 1;
		while(1) {
			if(board[i][j] == 0) {
				board[i][j] = 2;
				break;
			}
			switch(rand_int(3)) {
				case 0:
					i++;
					break;
				case 1:
					i--;
					break;
				default:
					j--;
					break;
			}
			if(i>col -1) i--;
			if(i< 0) i++;
		}
	}
}


void
green_draw(void)
{
	int i,j;
	for(i=0;i<col;i++) for(j=0;j<row;j++) 
		if(board[i][j]) draw_board(i,j);
}

int
check_collision(WordInfo *w)
{
	int collided = FALSE;
	int x = w->x;
	int y = w->y + 1;
	int i;
	if(y == row) return TRUE;
	for(i=x;i<x+w->size;i++) {
		if(board[i][y] != 0) {
			board[i][y]--;
			if(board[i][y] == 0) n_remaining_green--;
			draw_board(i, y);
			collided = TRUE;
		}
	}
	return collided;
}

void
draw_board(int x, int y)
{
	if(board[x][y] == 2) 
		gdk_draw_rectangle(drawing->window, green_gc,
			TRUE, x*font_width, y*font_height, font_width, font_height);
	else if(board[x][y] == 1)
		gdk_draw_rectangle(drawing->window, red_gc,
			TRUE, x*font_width, y*font_height, font_width, font_height);
	else 
		gdk_window_clear_area(drawing->window,
			x*font_width, y*font_height, font_width, font_height);
}

void level_callback(GtkWidget *w, gpointer user_data);
void destroy_callback(GtkWidget *w, gpointer user_data);
void level_destroy_callback(GtkWidget *w, gpointer user_data);
void continue_callback(GtkWidget *, GtkWidget *);
void popup_center(GtkWidget *w, GtkWidget *parent);


static GtkWidget *popup = NULL;

void
choose_level(void)
{
	GtkWidget *label, *vbox, *button, *separator, *frame;
	GSList *group;
	if(word_list) {
		g_list_free(word_list);
		word_list = NULL;
	}
	if(popup != NULL) {
		gdk_window_clear_area(drawing->window, 0, 0, col*font_width,
			row*font_height);
	}

	if(popup  == NULL) {
		popup = gtk_window_new(GTK_WINDOW_DIALOG);
		gtk_window_set_title(GTK_WINDOW(popup), " "); 
		gtk_container_border_width(GTK_CONTAINER(popup), 5);
		/*
		gtk_window_position(GTK_WINDOW(popup), GTK_WIN_POS_MOUSE);
		*/
		gtk_signal_connect(GTK_OBJECT(popup), "destroy",
			GTK_SIGNAL_FUNC(level_destroy_callback), NULL);
		
		vbox = gtk_vbox_new(FALSE, 3);
		gtk_container_add(GTK_CONTAINER(popup), vbox);
		gtk_widget_show(vbox);
		frame = gtk_frame_new(NULL);
		gtk_container_add(GTK_CONTAINER(vbox), frame);
		gtk_widget_show(frame);

		label = gtk_label_new(" ܰ踦 ");
		gtk_container_add(GTK_CONTAINER(frame), label);
		gtk_widget_show(label);

		button = gtk_radio_button_new_with_label(NULL, "1 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)0);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);

		button = gtk_radio_button_new_with_label(gtk_radio_button_group(
			GTK_RADIO_BUTTON(button)) , "2 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)1);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);

		button = gtk_radio_button_new_with_label(gtk_radio_button_group(
			GTK_RADIO_BUTTON(button)) , "3 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)2);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);

		button = gtk_radio_button_new_with_label(gtk_radio_button_group(
			GTK_RADIO_BUTTON(button)) , "4 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)3);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);

		button = gtk_radio_button_new_with_label(gtk_radio_button_group(
			GTK_RADIO_BUTTON(button)) , "5 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)4);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);

		button = gtk_radio_button_new_with_label(gtk_radio_button_group(
			GTK_RADIO_BUTTON(button)) , "6 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)5);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);

		button = gtk_radio_button_new_with_label(gtk_radio_button_group(
			GTK_RADIO_BUTTON(button)) , "7 ܰ");
		gtk_object_set_user_data(GTK_OBJECT(button), (gpointer)6);
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_widget_show(button);
		group = gtk_radio_button_group(GTK_RADIO_BUTTON(button));

		separator = gtk_hseparator_new();
		gtk_container_add(GTK_CONTAINER(vbox), separator);
		gtk_widget_show(separator);

		button = gtk_button_new_with_label("սô");
		gtk_container_add(GTK_CONTAINER(vbox), button);
		gtk_object_set_user_data(GTK_OBJECT(button), group);
		gtk_signal_connect(GTK_OBJECT(button), "clicked",
			GTK_SIGNAL_FUNC(level_callback), NULL);
		gtk_widget_show(button);
	}
	popup_center(popup, drawing);
	green_make();
}

void
level_callback(GtkWidget *w, gpointer user_data)
{
	extern int level;
	GSList *group = gtk_object_get_user_data(GTK_OBJECT(w));
	level = -1;
	gtk_widget_hide(popup);
	while(group) {
		GtkToggleButton *button = group->data;
		if(button->active) {
			level = (int)gtk_object_get_user_data(GTK_OBJECT(button));
			break;
		}
		group = group->next;
	}
	if(level == -1) level = 0;
	score = 0;
	n_appear_word = 0;
	n_failed_word = 0;
	n_destroyed_word = 0;
	
	/*
	gtk_widget_draw(drawing, NULL);
	*/
	begin_game();
}

void
level_destroy_callback(GtkWidget *w, gpointer user_data)
{
	popup = NULL;
}

void
end_dialog(void)
{
	GtkWidget *popup;
	GtkWidget *label, *button, *hbox, *vbox, *separator;
	popup = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_border_width(GTK_CONTAINER(popup), 5);
	gtk_window_set_title(GTK_WINDOW(popup), "ϴ");
	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(popup), vbox);
	gtk_widget_show(vbox);

	/*
	gtk_window_position(GTK_WINDOW(popup), GTK_WIN_POS_MOUSE);
	*/
	label = gtk_label_new(" ϴ");
	gtk_container_add(GTK_CONTAINER(vbox), label);
	gtk_widget_show(label);
	label = gtk_label_new("  Ͻðϱ?");
	gtk_container_add(GTK_CONTAINER(vbox), label);
	gtk_widget_show(label);

	separator = gtk_hseparator_new();
	gtk_container_add(GTK_CONTAINER(vbox), separator);
	gtk_widget_show(separator);

	hbox = gtk_hbox_new(TRUE, 5);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label(" Ҷ");
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(continue_callback), popup);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("׸ Ҷ");
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_widget_show(button);

	popup_center(popup, drawing);
	/*
	gtk_widget_show(popup);
	*/
}

void
popup_center(GtkWidget *w, GtkWidget *parent)
{
	int x, y;
	int width, height;
	int p_x, p_y;
	int p_w, p_h;
	if(!GTK_WIDGET_REALIZED(w)) gtk_widget_realize(w);
	gtk_widget_size_request(w, &w->requisition);
	width = w->requisition.width;
	height = w->requisition.height;
	gdk_window_get_size(parent->window, &p_w, &p_h);
	gdk_window_get_origin(parent->window, &p_x, &p_y);
	x = (p_x + p_w/2) - width/2;
	y = (p_y + p_h/2) - height/2;
	gtk_widget_popup(w, x, y);
}

void
continue_callback(GtkWidget *w, GtkWidget *popup)
{
	gtk_widget_destroy(popup);

	choose_level();
}

void
new_level_ok_callback(GtkWidget *w, GtkWidget *popup)
{
	gtk_widget_hide(popup);
	begin_game();
}

void
inform_new_level(int level)
{
	static GtkWidget *popup;
	GtkWidget *vbox;
	GtkWidget *label, *separator, *button;
	char s[60];
	if(popup) {
		gtk_widget_show(popup);
		return;
	}
	popup = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_border_width(GTK_CONTAINER(popup), 10);
    vbox  = gtk_vbox_new(FALSE, 5);
	gtk_container_add(GTK_CONTAINER(popup), vbox);
	gtk_widget_show(vbox);
	sprintf(s, "%d ܰ踦 մϴ", level + 1);
	label = gtk_label_new(s);
	gtk_container_add(GTK_CONTAINER(vbox), label);
	gtk_widget_show(label);

	separator = gtk_hseparator_new();
	gtk_container_add(GTK_CONTAINER(vbox), separator);
	gtk_widget_show(separator);

	button = gtk_button_new_with_label("׷ô");
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		GTK_SIGNAL_FUNC(new_level_ok_callback), popup);
	gtk_container_add(GTK_CONTAINER(vbox), button);
	
	gtk_widget_show(button);
	popup_center(popup, drawing);
}


int count_space(char *s);


int
word_data_parse(void)
{
	int i, wordsize;
	char *w;
	wordsize = strlen(words);
	nword = count_space(words);
	word = g_new(char * , nword);
	w = words;
	for(i=0;i<nword;i++) {
		word[i] = w;
		while(*w && !isspace(*w) ) w++;
		*w++ = 0;
	}
	return 1;
}

int
count_space(char *s)
{
	int count = 0;
	while(*s) {
		if(isspace(*s++)) count++;
	}
	return count;
}

