/*

 gzilla

 Copyright 1997 Raph Levien <raph@acm.org>

 This code is free for commercial and non-commercial use,
 modification, and redistribution, as long as the source code release,
 startup screen, or product packaging includes this copyright notice.

 */

#include <gtk/gtk.h>
#include "gzillabytesink.h"

enum {
  WRITE,
  CLOSE,
  RESET,
  REQUEST_URL,
  SET_BASE_URL,
  LINK,
  STATUS,
  REDIRECT,
  TITLE,
  REQUEST_URL_IMG,
  LAST_SIGNAL
};

typedef void (*GzillaByteSinkSignal1) (GtkObject *object,
				       gpointer  arg1,
				       gint      arg2,
				       gpointer  data);

typedef void (*GzillaByteSinkSignal2) (GtkObject *object,
				       gpointer  arg1,
				       gpointer  arg2,
				       gpointer  data);

typedef void (*GzillaByteSinkSignal3) (GtkObject *object,
				       gpointer  arg1,
				       gpointer  data);

static void gzilla_bytesink_marshal_signal_1 (GtkObject     *object,
					      GtkSignalFunc  func,
					      gpointer       func_data,
					      GtkArg        *args);
static void gzilla_bytesink_marshal_signal_2 (GtkObject     *object,
					      GtkSignalFunc  func,
					      gpointer       func_data,
					      GtkArg        *args);
static void gzilla_bytesink_marshal_signal_3 (GtkObject     *object,
					      GtkSignalFunc  func,
					      gpointer       func_data,
					      GtkArg        *args);

static void gzilla_bytesink_class_init  (GzillaByteSinkClass  *klass);
static void gzilla_bytesink_init        (GzillaByteSink       *widget);

static GtkObjectClass *parent_class = NULL;
static gint bytesink_signals[LAST_SIGNAL] = { 0 };

guint gzilla_bytesink_get_type () {
  static guint bytesink_type = 0;

  if (!bytesink_type)
    {
      GtkTypeInfo bytesink_info =
      {
	"GzillaByteSink",
	sizeof (GzillaByteSink),
	sizeof (GzillaByteSinkClass),
	(GtkClassInitFunc) gzilla_bytesink_class_init,
	(GtkObjectInitFunc) gzilla_bytesink_init,
	(GtkArgFunc) NULL
      };

      bytesink_type = gtk_type_unique (gtk_object_get_type (), &bytesink_info);
    }

  return bytesink_type;
}

static void gzilla_bytesink_class_init (GzillaByteSinkClass *klass)
{
  GtkObjectClass *object_class;
  gint i;

  object_class = (GtkObjectClass *)klass;

  parent_class = gtk_type_class (gtk_object_get_type ());

  i = 0;
  bytesink_signals[i++] =
    gtk_signal_new ("write",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, write),
		    gzilla_bytesink_marshal_signal_1, GTK_ARG_NONE, 2,
		    GTK_ARG_POINTER,
		    GTK_ARG_INT);
  bytesink_signals[i++] =
    gtk_signal_new ("close",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, close),
		    gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  bytesink_signals[i++] =
    gtk_signal_new ("reset",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, reset),
		    gtk_signal_default_marshaller, GTK_ARG_NONE, 0);
  bytesink_signals[i++] =
    gtk_signal_new ("request_url",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, request_url),
		    gzilla_bytesink_marshal_signal_2, GTK_ARG_NONE, 2,
		    GTK_ARG_POINTER,
		    GTK_ARG_POINTER);
  bytesink_signals[i++] =
    gtk_signal_new ("set_base_url",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, set_base_url),
		    gzilla_bytesink_marshal_signal_3, GTK_ARG_NONE, 1,
		    GTK_ARG_POINTER);
  bytesink_signals[i++] =
    gtk_signal_new ("link",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, link),
		    gzilla_bytesink_marshal_signal_3, GTK_ARG_NONE, 1,
		    GTK_ARG_POINTER);
  bytesink_signals[i++] =
    gtk_signal_new ("status",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, status),
		    gzilla_bytesink_marshal_signal_3, GTK_ARG_NONE, 1,
		    GTK_ARG_POINTER);
  bytesink_signals[i++] =
    gtk_signal_new ("redirect",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, redirect),
		    gzilla_bytesink_marshal_signal_3, GTK_ARG_NONE, 1,
		    GTK_ARG_POINTER);
  bytesink_signals[i++] =
    gtk_signal_new ("title",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, title),
		    gzilla_bytesink_marshal_signal_3, GTK_ARG_NONE, 1,
		    GTK_ARG_POINTER);
  bytesink_signals[i++] =
    gtk_signal_new ("request_url_img",
		    GTK_RUN_FIRST,
		    object_class->type,
		    GTK_SIGNAL_OFFSET (GzillaByteSinkClass, request_url_img),
		    gzilla_bytesink_marshal_signal_2, GTK_ARG_NONE, 2,
		    GTK_ARG_POINTER,
		    GTK_ARG_POINTER);

  gtk_object_class_add_signals (object_class, bytesink_signals, LAST_SIGNAL);

#if 0
  /* we should probably destroy the associated widget */
  object->class_destroy = gzilla_bytesink_real_bytesink_destroy;
#endif

  klass->write = NULL;
  klass->close = NULL;
  klass->reset = NULL;
  klass->request_url = NULL;
  klass->set_base_url = NULL;
  klass->link = NULL;
  klass->status = NULL;
  klass->redirect = NULL;
  klass->title = NULL;
  klass->request_url_img = NULL;
}

static void gzilla_bytesink_init (GzillaByteSink *bytesink) {
  bytesink->widget = NULL;
}

void gzilla_bytesink_write (GzillaByteSink *bytesink,
			    char *buf,
			    gint bufsize) {
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[WRITE],
		   buf, bufsize);
}

void gzilla_bytesink_close (GzillaByteSink *bytesink) {
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[CLOSE]);
}

void gzilla_bytesink_reset (GzillaByteSink *bytesink) {
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[RESET]);
}

void
gzilla_bytesink_set_base_url (GzillaByteSink *bytesink,
			      const char *url)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[SET_BASE_URL],
		   url);
}

void
gzilla_bytesink_request_url (GzillaByteSink *bytesink,
			     GzillaByteSink *sink2,
			     const char *url)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[REQUEST_URL],
		   sink2, url);
}

void
gzilla_bytesink_link (GzillaByteSink *bytesink,
			   const char *url)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[LINK],
		   url);
}

void
gzilla_bytesink_status (GzillaByteSink *bytesink,
			const char *string)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[STATUS],
		   string);
}

void
gzilla_bytesink_redirect (GzillaByteSink *bytesink,
			  const char *url)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[REDIRECT],
		   url);
}

void
gzilla_bytesink_title (GzillaByteSink *bytesink,
		       const char *title)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[TITLE],
		   title);
}

void
gzilla_bytesink_request_url_img (GzillaByteSink *bytesink,
				 GzillaImgSink *imgsink,
				 const char *url)
{
  gtk_signal_emit (GTK_OBJECT (bytesink), bytesink_signals[REQUEST_URL_IMG],
		   imgsink, url);
}

static void
gzilla_bytesink_marshal_signal_1 (GtkObject      *object,
				  GtkSignalFunc   func,
				  gpointer        func_data,
				  GtkArg         *args)
{
  GzillaByteSinkSignal1 rfunc;

  rfunc = (GzillaByteSinkSignal1) func;

  (* rfunc) (object, args[0].d.p, args[1].d.i, func_data);
}

static void
gzilla_bytesink_marshal_signal_2 (GtkObject      *object,
				  GtkSignalFunc   func,
				  gpointer        func_data,
				  GtkArg         *args)
{
  GzillaByteSinkSignal2 rfunc;

  rfunc = (GzillaByteSinkSignal2) func;

  (* rfunc) (object, args[0].d.p, args[1].d.p, func_data);
}

static void
gzilla_bytesink_marshal_signal_3 (GtkObject      *object,
				  GtkSignalFunc   func,
				  gpointer        func_data,
				  GtkArg         *args)
{
  GzillaByteSinkSignal3 rfunc;

  rfunc = (GzillaByteSinkSignal3) func;

  (* rfunc) (object, args[0].d.p, func_data);
}
