#include <glib.h>
#include <gtk/gtkcontainer.h>
#include <stdio.h>
#include <string.h>

#include "phone.h"

static void phone_class_init    (PhoneClass *klass);
static void phone_init          (Phone      *darea);
static void phone_realize       (GtkWidget           *widget);
static void phone_destroy       (GtkObject *object);
static void phone_size_allocate (GtkWidget           *widget,
					    GtkAllocation       *allocation);
static void phone_size_request (GtkWidget           *widget,
				      GtkRequisition *requisition);
static gint phone_expose          (GtkWidget      *widget,
				      GdkEventExpose *event);
static void redraw_nth_item(Phone *phone, int position);
static gint phone_button_press          (GtkWidget      *widget,
				      GdkEventButton *event);

void phone_item_destroy(PhoneItem *item);

static GtkWidgetClass *parent_class = NULL;
static GdkPixmap *selected_pixmap = NULL;
static int selected_pixmap_width = 25 ;

guint
phone_get_type ()
{
  static guint phone_type = 0;

  if (!phone_type)
    {
      GtkTypeInfo phone_info =
      {
	"Phone",
	sizeof (Phone),
	sizeof (PhoneClass),
	(GtkClassInitFunc) phone_class_init,
	(GtkObjectInitFunc) phone_init,
	(GtkArgFunc) NULL,
      };

      phone_type = gtk_type_unique (gtk_widget_get_type (), &phone_info);
    }
  // g_print("%s in %s\n", __FUNCTION__, __FILE__);

  return phone_type;
}

#include "checked.xpm"
static void
phone_class_init (PhoneClass *class)
{
	GtkWidgetClass *widget_class;
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass*) class;
	widget_class = (GtkWidgetClass*) class;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);

	parent_class = gtk_type_class (gtk_widget_get_type ());
	object_class->destroy = phone_destroy;
	widget_class->realize = phone_realize;
	widget_class->size_allocate = phone_size_allocate;
	widget_class->size_request = phone_size_request;
	widget_class->expose_event = phone_expose;
	widget_class->button_press_event = phone_button_press;
}

static void
phone_init (Phone *phone)
{
	GTK_WIDGET_SET_FLAGS(phone, GTK_CAN_FOCUS);
	phone->items = NULL;
	phone->nitem = 0;
	phone->draw_flag = TRUE;
	phone->height = 0;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
}


GtkWidget*
phone_new (PhoneLayout *dis, char **title)
{
	Phone *phone = gtk_type_new (phone_get_type ());
	float total;
	int i;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
	dis->total_size -= selected_pixmap_width;
	phone->layout = g_new(PhoneLayout, 1);
	phone->layout->number = dis->number;
	phone->layout->total_size = dis->total_size;
	phone->layout->size = g_new(int, dis->number);
	GTK_WIDGET(phone)->allocation.width = dis->total_size + selected_pixmap_width;
	GTK_WIDGET(phone)->allocation.height = 100;
	total = 0.0;
	for(i=0;i<dis->number;i++) {
		total += dis->size[i];
	}
	for(i=0;i<dis->number;i++) {
		phone->layout->size[i] = dis->total_size*dis->size[i]/total;
		//g_print("%d %d\n", i, phone->layout->size[i]);
	}
	for(i=dis->number -1;i>=0;i--) {
		int j;
		phone->layout->size[i] = 0;
		for(j=i-1;j>=0;j--) {
			phone->layout->size[i] += phone->layout->size[j];
		}
		phone->layout->size[i] += selected_pixmap_width;
	}
	phone->title = g_new(char *, dis->number);
	for(i=0;i<dis->number;i++) {
		phone->title[i] = g_strdup(title[i]);
	}
	return GTK_WIDGET(phone);
}

static void
phone_realize (GtkWidget *widget)
{
	Phone *darea;
	GdkWindowAttr attributes;
	gint attributes_mask;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (widget != NULL);


  darea = PHONE (widget);
  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);
  attributes.event_mask = gtk_widget_get_events (widget);
  attributes.event_mask |= (GDK_EXPOSURE_MASK |
  	GDK_BUTTON_PRESS_MASK |
	GDK_KEY_PRESS_MASK);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);
  gdk_window_set_user_data (widget->window, darea);

  widget->style = gtk_style_attach (widget->style, widget->window);
  gdk_window_set_background(widget->window, &widget->style->white);

	if(selected_pixmap == NULL) {
		GdkColor *bg = &widget->style->bg[GTK_STATE_NORMAL];
		selected_pixmap = gdk_pixmap_create_from_xpm_d(NULL, NULL, bg,
		checked_data);
	}
}

static void
phone_size_allocate (GtkWidget     *widget,
				GtkAllocation *allocation)
{
  //g_print("%s in %s\n", __FUNCTION__, __FILE__);
  g_return_if_fail (widget != NULL);
  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED (widget))
    {
      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

    }
}

void
phone_insert_item(Phone *phone, PhoneItem *item, int position)
{
	int i;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (phone != NULL);
	if(phone->nitem == 0) {
		phone->current = 0;
		item->state |= PHONE_CURRENT;
	} else 
		item->state &= ~PHONE_CURRENT;
	phone->items = g_list_insert(phone->items, item, position);
	//g_print("%d %d => ", phone->current, position);
	if(phone->current >= position && phone->nitem) phone->current++;
	//g_print("%d %d \n", phone->current, position);
	phone->nitem++;

	//printf("%d %d %d\n", phone->nitem, phone->height, GTK_WIDGET(phone)->allocation.height);
	if(phone->draw_flag) {
		if(phone->nitem * phone->height > GTK_WIDGET(phone)->allocation.height) {
			gtk_container_check_resize (GTK_CONTAINER 
				(GTK_WIDGET (phone)->parent), GTK_WIDGET (phone));
		} else  
			for(i=0;i<phone->nitem;i++) redraw_nth_item(phone, i);
	}

}

void
phone_append_item(Phone *phone, PhoneItem *item)
{
	phone_insert_item(phone, item, phone->nitem);
}

void
phone_prepend_item(Phone *phone, PhoneItem *item)
{
	phone_insert_item(phone, item, 0);
}

void
phone_remove_nth_item(Phone *phone, int position)
{
	PhoneItem *item;
	GList *list;
	int i;
	//g_print("%s in %s %d\n", __FUNCTION__, __FILE__, position);
	g_return_if_fail (phone != NULL);
	list = g_list_nth(phone->items, position);
	//g_print("%s in %s %p\n", __FUNCTION__, __FILE__, list);
	if(list) {
		//g_print("%s in %s %d\n", __FUNCTION__, __FILE__, phone->nitem);
		item = list->data;
		if(position == phone->current) {
			if(list->next) {
				((PhoneItem *)list->next->data)->state |= PHONE_CURRENT;
			} else if(list->prev) {
				((PhoneItem *)list->prev->data)->state |= PHONE_CURRENT;
				phone->current--;
			}
		}
		phone_item_destroy(item);
		phone->items = g_list_remove_link(phone->items, list);
		phone->nitem--;
		if(phone->draw_flag) {
			gtk_container_check_resize (GTK_CONTAINER 
				(GTK_WIDGET (phone)->parent), GTK_WIDGET (phone));
			gdk_window_clear_area(GTK_WIDGET(phone)->window, 0, 
				phone->height*(phone->nitem + 1),
				phone->layout->total_size + selected_pixmap_width,
				phone->height);
			for(i=position-1;i<phone->nitem;i++) 
				redraw_nth_item(phone, i);
		}
	}
}

static gint
phone_expose(GtkWidget *w, GdkEventExpose *ev)
{
	GdkFont *font;
	int height;
	int index1;
	int index2;
	int i, index;
	Phone *phone;
	g_return_val_if_fail (w != NULL, FALSE);

	phone = PHONE(w);
	font = w->style->font;
	height = phone->height;
	index1 = ev->area.y/height;
	index2 = (ev->area.y + ev->area.height)/height  - 1;
	if(index1 == 0) {
		int y = 0;
		gdk_draw_rectangle(w->window, w->style->bg_gc[GTK_STATE_INSENSITIVE],
			TRUE, 0, 0, phone->layout->total_size + selected_pixmap_width, height); 
		for(i=0;i<phone->layout->number;i++) {
			gdk_draw_string(w->window, w->style->white_gc,
				phone->layout->size[i] + 4, font->ascent + 2, phone->title[i]);
			gdk_draw_string(w->window, w->style->black_gc,
				phone->layout->size[i] + 3, font->ascent + 1, phone->title[i]);
		}
		for(i=0;i<phone->layout->number;i++) {
			gtk_draw_vline(w->style, w->window, GTK_STATE_NORMAL,
				y, y + phone->height, phone->layout->size[i]);
		}
		gtk_draw_vline(w->style, w->window, GTK_STATE_NORMAL,
			y, y + phone->height, phone->layout->total_size + selected_pixmap_width);
		index1++;
	}
	index1--;

	if(index2 >= phone->nitem) index2 = phone->nitem -1;
	for(index=index1;index<=index2;index++) {
		redraw_nth_item(phone, index);
	}
	// g_print("%s in %s\n", __FUNCTION__, __FILE__);
	return FALSE;
}

static void
redraw_nth_item(Phone *phone, int position)
{
	int y;
	GtkWidget *w = GTK_WIDGET(phone);
	int i;
	GdkGC *gc;
	PhoneItem *phoneinfo;
	if(position < 0 || position >= phone->nitem) return;
	y = phone->height * (position + 1);

	phoneinfo = g_list_nth(phone->items, position)->data;
	if(phoneinfo->state & PHONE_SELECTED) {
		gdk_draw_pixmap(w->window, w->style->black_gc, selected_pixmap,
			0, 0, 0, y, -1, -1);
	} else {
		gdk_draw_rectangle(w->window, w->style->bg_gc[GTK_STATE_INSENSITIVE],
			TRUE, 0, y, selected_pixmap_width, phone->height); 
	}
	if(phoneinfo->state & PHONE_CURRENT) {
		gdk_draw_rectangle(w->window, w->style->bg_gc[GTK_STATE_SELECTED],
			TRUE, selected_pixmap_width, y + 1, 
			phone->layout->total_size, phone->height -2);
		gc = w->style->fg_gc[GTK_STATE_SELECTED];
	} else {
		gc = w->style->black_gc;
		gdk_window_clear_area(w->window, 
			selected_pixmap_width, y, 
			phone->layout->total_size, phone->height);
	}


	if(phoneinfo->name)
	gdk_draw_string(w->window, gc,
		phone->layout->size[0] + 4, y + w->style->font->ascent + 1, 
		phoneinfo->name);
	if(phoneinfo->phone_number)
	gdk_draw_string(w->window, gc,
		phone->layout->size[1] + 4, y + w->style->font->ascent + 1, 
		phoneinfo->phone_number);
	if(phoneinfo->speed)
	gdk_draw_string(w->window, gc,
		phone->layout->size[2] + 4, y + w->style->font->ascent + 1, 
		phoneinfo->speed);
	if(phoneinfo->script_file)
	gdk_draw_string(w->window, gc,
		phone->layout->size[3] + 4, y + w->style->font->ascent + 1, 
		phoneinfo->script_file);

	for(i=0;i<phone->layout->number;i++) {
	}
	/* border lines */
	gtk_draw_hline(w->style, w->window, GTK_STATE_NORMAL,
		0, phone->layout->total_size + selected_pixmap_width, y);
	gtk_draw_hline(w->style, w->window, GTK_STATE_NORMAL,
		0, phone->layout->total_size + selected_pixmap_width, y + phone->height -1);
	for(i=0;i<phone->layout->number;i++) {
		gtk_draw_vline(w->style, w->window, GTK_STATE_INSENSITIVE,
			y, y + phone->height, phone->layout->size[i]);
	}
	gtk_draw_vline(w->style, w->window, GTK_STATE_INSENSITIVE,
		y, y + phone->height, phone->layout->total_size + selected_pixmap_width);
}

static gint
phone_button_press(GtkWidget *w, GdkEventButton *ev)
{
	Phone *phone;
	int y;
	PhoneItem *item;
	phone = PHONE(w);
	y = ev->y/phone->height;
	if(y == 0) return FALSE;
	y--;
	if(y >= PHONE(w)->nitem) return FALSE;
	if(ev->type == GDK_BUTTON_PRESS) {
		item = g_list_nth(phone->items, y)->data;
		item->state ^= PHONE_SELECTED;
		item->state |= PHONE_CURRENT;
		redraw_nth_item(phone, y);
		if(y != phone->current) {
			item = g_list_nth(phone->items, phone->current)->data;
			item->state &= ~PHONE_CURRENT;
			redraw_nth_item(phone, phone->current);
		}
		phone->current = y;
	}

	//g_print("%s\n", __FUNCTION__);
	return FALSE;
}

void
phone_item_destroy(PhoneItem *item)
{
	extern void destroy_script(GList *);
	if(item->name) g_free(item->name);
	if(item->phone_number) g_free(item->phone_number);
	if(item->speed) g_free(item->speed);
	if(item->script_file) g_free(item->script_file);
	g_free(item);
}

static void
phone_destroy(GtkObject *object)
{
	Phone *phone = PHONE(object);
	int number = phone->layout->number; 
	GList *list = phone->items;
	int i;
	//g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (object != NULL);
	for(i=0;i<number;i++) g_free(phone->title[i]);
	g_free(phone->title);
	g_free(phone->layout->size);
	g_free(phone->layout);
	while(list) {
		g_free(list->data);
		list = list->next;
	}
	g_list_free(phone->items);

	if (GTK_OBJECT_CLASS (parent_class)->destroy)
		(* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

static void
phone_size_request(GtkWidget *widget, GtkRequisition *requisition)
{
  //g_print("%s in %s\n", __FUNCTION__, __FILE__);
	g_return_if_fail (widget != NULL);
	if(PHONE(widget)->height == 0) {
		PHONE(widget)->height = 
			widget->style->font->ascent + widget->style->font->descent;
		if(PHONE(widget)->height < 22) PHONE(widget)->height = 22;
	}
	requisition->width = widget->allocation.width;
	requisition->height = PHONE(widget)->height * (PHONE(widget)->nitem + 1);
}

GtkWidget *
phone_new_from_file(char *filename)
{
	GtkWidget *widget;
	FILE *fp;
	int number;
	PhoneLayout *layout;
	PhoneItem *item;
	int total_size;
	int *size;
	int i;
	char **title, buf[50], *str, *s;

	fp = fopen(filename, "r");
	if(fp == NULL) return NULL;
	fscanf(fp, "%d\n", &number);
	//g_print("number = %d\n", number);
	fscanf(fp, "%d\n", &total_size);
	//g_print("total_size = %d\n", total_size);
	size = g_new(int, number);
	for(i=0;i<number -1;i++) fscanf(fp, "%d ", size + i);
	fscanf(fp, "%d\n", size + i);
	layout = g_new(PhoneLayout, 1);
	layout->size = size;
	layout->number = number;
	layout->total_size = total_size;
	title = g_new(char *, number);
	for(i=0;i<number;i++) {
		fscanf(fp, "%s\n", buf);
		//g_print("'%s'\n", buf);
		title[i] = g_strdup(buf);
	}
	widget = phone_new(layout, title);
	phone_set_draw_flag(PHONE(widget), FALSE);
	while(fscanf(fp, "%s", buf) != EOF) {
		item = g_new0(PhoneItem, 1);
		item->state = 0;
		str = buf;
		s = strchr(str, ';');
		*s = 0;
		item->name = g_strdup(str);
		str = s+1;
		s = strchr(str, ';');
		*s = 0;
		item->phone_number = g_strdup(str);
		str = s+1;
		s = strchr(str, ';');
		*s = 0;
		item->speed = g_strdup(str);
		str = s+1;
		s = strchr(str, ';');
		*s = 0;
		if(*str == 0) item->script_file = NULL;
		else item->script_file = g_strdup(str);
		phone_append_item(PHONE(widget), item);
	}

	for(i=0;i<number;i++) {
		g_free(title[i]);
	}
	g_free(title);
	g_free(size);
	g_free(layout);
	phone_set_draw_flag(PHONE(widget), TRUE);
	return widget;
}


int
phone_file_save(Phone *phone, char * filename)
{
	FILE *fp;
	PhoneItem *item;
	GList *list;
	int i;
	
	fp = fopen(filename, "w");
	if(fp == NULL) return FALSE;

	/* save layout */
	fprintf(fp, "%d\n", phone->layout->number);
	fprintf(fp, "%d\n", phone->layout->total_size + selected_pixmap_width);
	for(i=0;i<phone->layout->number -1;i++) {
		fprintf(fp, "%d ", phone->layout->size[i+1] - phone->layout->size[i]);
	}
	fprintf(fp, "%d\n", phone->layout->total_size + selected_pixmap_width - phone->layout->size[i]);
	for(i=0;i<phone->layout->number;i++) {
		fprintf(fp, "%s\n", phone->title[i]);
	}

	/* save data */
	list = phone->items;
	while(list){
		item = list->data;
		if(item->name) fprintf(fp, "%s;", item->name);
		else fprintf(fp, ";");
		if(item->phone_number) fprintf(fp, "%s;", item->phone_number);
		else fprintf(fp, ";");
		if(item->speed) fprintf(fp, "%s;", item->speed);
		else fprintf(fp, ";");
		if(item->script_file) fprintf(fp, "%s;", item->script_file);
		else fprintf(fp, ";");
		fprintf(fp, "\n");
		list = list->next;
	}
	fclose(fp);
	return TRUE;
}

void
phone_set_draw_flag(Phone *phone, int flag)
{
	phone->draw_flag = flag;
}

void
phone_selection_toggle(Phone *phone, int position)
{
	((PhoneItem *)g_list_nth(phone->items, position)->data)->state ^= PHONE_SELECTED;
	redraw_nth_item(phone, position);
}
