/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef __PHONE_H__
#define __PHONE_H__


#include <gdk/gdk.h>
#include <gtk/gtkwidget.h>


#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define PHONE_SELECTED 1
#define PHONE_CURRENT 2

#define PHONE(obj)          GTK_CHECK_CAST (obj, phone_get_type (), Phone)
#define GTK_PHONE_CLASS(klass)  GTK_CHECK_CLASS_CAST (klass, phone_get_type (), GtkDrawingAreaClass)
#define GTK_IS_PHONE(obj)       GTK_CHECK_TYPE (obj, phone_get_type ())


typedef struct _Phone       Phone;
typedef struct _PhoneClass  PhoneClass;
typedef struct _PhoneLayout PhoneLayout;
typedef struct _PhoneItem   PhoneItem;

struct _Phone
{
  GtkWidget widget;
  PhoneLayout *layout;
  char **title;
  GList *items;
  int current;
  int nitem;
  int height;
  int draw_flag;
};

struct _PhoneItem
{
	int state;
	char *speed;
	char *phone_number;
	char *name;
	char *script_file;
};

struct _PhoneClass
{
  GtkWidgetClass parent_class;
};

struct _PhoneLayout
{
	int number;
	int total_size;
	int *size;
};


guint      phone_get_type   (void);
GtkWidget *phone_new (PhoneLayout *dis, char **title);
GtkWidget *phone_new_from_file (char *);
int  phone_file_save(Phone *phone, char *);
void phone_insert_item(Phone *phone, PhoneItem *item, int position);
void phone_append_item(Phone *phone, PhoneItem *item);
void phone_prepend_item(Phone *phone, PhoneItem *item);
void phone_remove_nth_item(Phone *phone, int position);
void phone_set_draw_flag(Phone *phone, int flat);
void phone_selection_toggle(Phone *phone, int position);

#ifdef __cplusplus
}
#endif /* __cplusplus */


#endif /* __PHONE_H__ */
