
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<stdio.h>
#include	<signal.h>
#include    <unistd.h>
#include    <sys/ioctl.h>
#include    <string.h>

/* this file is adapted from splitvt. */


#include	<termios.h>

/*
 * Initialize a pty, fork a command running under it, and then 
 * return the master file descriptor
 */


void dropctty(void);
int tty_reset(int fd);
int tty_raw(int fd);


int pty_open(command, childpid, col, row)
char *command;
int *childpid;
int col, row;
{

	void dropctty(), pty_setwinsize();
	int get_master_pty(), get_slave_pty();
	char *get_ttyname();

	int returnfd, slave_fd;

	/* Get the master pty file descriptor */
	if ( (returnfd=get_master_pty()) < 0 )
		return(-1);

	/* Fork and go! */
	if ( ((*childpid)=fork()) < 0 )
		return(-1);
	else if ( (*childpid) == 0 )
	{
		dropctty();	/* Lose controlling tty */

		if ( (slave_fd=get_slave_pty()) < 0 )
		{
			perror("Can't open slave tty");
			exit(128);
		}
		close(0); close(1); close(2);
		dup(slave_fd); dup(slave_fd); dup(slave_fd);
		close(slave_fd); close(returnfd);

		/* Reattatch the new tty as the controlling terminal */
		/* Under old UNIX, just opening the new tty after
		   losing the controlling tty is good enough.
		   Under newer Unices, it requires an ioctl().  */
		(void) ioctl(0, TIOCSCTTY, 0);

		/* Set the lines and columns on the new tty */
		pty_setwinsize(0, col, row);
		signal(SIGTSTP, SIG_IGN);
		//setbuf(stdout, NULL);
		//tty_raw(1);

		/* Set our uid to our real uid if necessary */
		(void) setuid(getuid());
			
		/* Run the requested program, with possible leading dash. */
		execlp("sh", "sh", "-c", command, NULL);
		_exit(1);
	}
	return(returnfd);
}



/*
 * Pseudo-terminal routines for Unix System V Release 3.2 and BSD4.2-3
 */


extern int errno;
int master_fd;

char tty_name[64]={'\0'};
char pty_name[64]={'\0'};

char *get_ttyname()
{
	if ( tty_name[0] )
		return(tty_name);
	return(NULL);
}

#include	<setjmp.h>

#ifndef X_OK
#define	R_OK	4	/* Test for Read permission */
#define	W_OK	2	/* Test for Write permission */
#define	X_OK	1	/* Test for eXecute permission */
#endif  /* X_OK */

jmp_buf next;

void trynext()
{
	longjmp(next, 2);
}


int get_master_pty()
{
	int i, master_fd;
	char *ptr;
	struct stat statbuff;
#ifdef PTYCHAR
	static char ptychar[]=PTYCHAR;			/* X */ 
	static char hexdigit[]=HEXDIGIT;		/* Y */
#else
	static char ptychar[]="pqrstuvwxyzPQRST";	/* X */ 
	static char hexdigit[]="0123456789abcdef";	/* Y */
#endif

	for (ptr=ptychar; *ptr != 0; ptr++)
	{
		strcpy(pty_name, "/dev/ptyXY");
		pty_name[8]=(*ptr);  /* X */
		pty_name[9]='0';   /* Y */

		if ( stat(pty_name, &statbuff) < 0 )
			break;
#ifdef OLDDEBUG
		fprintf(stderr, "statted.\n");
#endif
		i=(-1);		/* Initialize i */

		/* Set a time limit for the open */
		if ( setjmp(next) == -1 )
			return(-1);
		signal(SIGALRM, trynext);

		for ( ++i; hexdigit[i]; ++i)
		{
			pty_name[5]='p';
			pty_name[9]=hexdigit[i];

			alarm(2);	/* Set an open timeout */

			if ( (master_fd=open(pty_name, O_RDWR)) >= 0 )
			{
				alarm(0); 	/* Reset the alarm */

				pty_name[5]='t';
				sprintf(tty_name, "%s", pty_name);
#ifdef OLDDEBUG
				fprintf(stderr, "tty: %s\n", tty_name);
#endif
				if ( access(tty_name, R_OK|W_OK) == 0 ) {
					signal(SIGALRM, SIG_DFL);
					return (master_fd);
				} else {
					pty_name[5]='p';
					(void) close(master_fd);
				}
			}
			/* reset the alarm */
			alarm(0);
		}
	}
	return(-1);
}


/* Open the slave half of a pseudo-terminal. */

int get_slave_pty()
{
	int slave_fd;

	if ( (slave_fd=open(tty_name, O_RDWR)) < 0 )
	{
		close(master_fd);
		return(-1);
	}
	return(slave_fd);
}


/* These are the binary data functions that I am using instead of 
   bcopy() and bzero(), written by Richard A. O'Keefe.
     Thanks!
*/


void d_zero(dst, len)
    register char *dst;
    register int len;
    {
	while (--len >= 0) *dst++ = 0;
    }



/* Here are the Terminal manipulation routines...  */


/* Code to disassociate from my tty. Yay! :) */

void dropctty()
{
	setsid();		/* The POSIX solution is simple. :) */
}



/* Get the modes of the controlling tty and save them.  Saves
   ttymodes in tty_mode and returns -1 if ioctl fails. */

struct termio tty_mode;  /* Save tty mode here */

/* Set a terminal in raw mode */

int
tty_raw(int fd)
{
	struct termios tty;
	if (!isatty(fd)) return 0;
	tcgetattr(fd, &tty);

	cfmakeraw(&tty);

	tcsetattr(fd, TCSANOW, &tty);
	return 0;
}

/* Set the pty window size to the size of the virtual window */



void pty_setwinsize(fd, col, row)
int fd;			/* The pty file descriptor */
int col, row;   // size of this terminal
{
	static struct winsize mywinz;
	mywinz.ws_row=row;
	mywinz.ws_col=col;
	mywinz.ws_xpixel=0;
	mywinz.ws_ypixel=0;
	(void) ioctl(fd, TIOCSWINSZ, &mywinz);
}
