/* -------------------------------------------------------------------------- */
/*                                                                            */
/*            ZMODEM File Transport                            */
/*                                                                            */
/* -------------------------------------------------------------------------- */
/*                                                                            */
/* Copyright (c) 1997 by Lars Doelle                                          */
/* Copyright (c) 1997 by Hwang Chiduck 										  */
/*                                                                            */
/* This file is part of gau - a serial line terminal                          */
/*                                                                            */
/* The whole program is available under the GNU Public Software Licence.      */
/* See COPYING, the documenation, or <http://www.gnu.org> for details.        */
/*                                                                            */
/* -------------------------------------------------------------------------- */

/* outline:
   - we first separate blocks
     - header
     - data packages
   - remainder can be controled by few state variables
*/

/* CURRENT STATE:
   - 70 % of zmodem receiver
*/
    
#include <stdio.h>
#include <sys/stat.h>
#include <ctype.h>
#include <unistd.h>
#include <gtk/gtk.h>

#include "ansiterm.h"
#include "gau.h"
#include "config.h"


#include "zmodem.h"
#include "zmodcon.h"
#include "crc.h"

void zmodem_dmy_newBlock(long size);
void zmodem_dmy_newError(void);
void zmodem_dmy_newFile(const char *, long);
void zmodem_dmy_doneFile(void);
void zmodem_dmy_doneTransmission(void);
void zmodem_done(void);
void zmodem_skip(void);
void zmodem_rcv_error(char* comment, guint8* buffer, int len);
void zmodem_rcv_fatal(char *);
void zmodem_rcv_cancel(void);
void zmodem_rcv_header(void);
void zmodem_receiver_rcv_error(void );
void zmodem_receiver_rcv_header(guint8, guint32);
void zmodem_receiver_rcv_frame(guint8, guint8, guint8 *, int);
FILE * new_file(char *filename);

#define MAXBUFLEN 3000
long int curr_file_offset;
FILE *curr_file;
char buffer[MAXBUFLEN];
char frame_buffer[3000];
int  buftop, zdlecnt, frame_buftop, last_hdr_typ;
int key_pos, expect00;
long    bytes_total;
long filesize; long tim; int mode; int nfilesleft; long totalleft;
long	bytes_skipped;

void 
hexdump(char* header, char* buffer, int len)
{
	int i;
  printf("%s\r\n  ",header);
  for (i = 0; i < len; i++)
  {
    if (i > 0 && i %  4 == 0) printf(" ");
    if (i > 0 && i %  8 == 0) printf(" ");
    if (i > 0 && i % 16 == 0) printf("\r\n  ");
    printf("%02x ",(guint8)buffer[i]);
  }
  printf("\r\n");
  fflush(stdout);
}

#if 0
static char *frametypes[] = 
{
	"ZRQINIT", "ZRINIT", "ZSINIT",   "ZACK",     "ZFILE",
	"ZSKIP",   "ZNAK",   "ZABORT",   "ZFIN",     "ZRPOS",
	"ZDATA",   "ZEOF",   "ZFERR",    "ZCRC",     "ZCHALLENGE",
	"ZCOMPL",  "ZCAN",   "ZFREECNT", "ZCOMMAND", "ZSTDERR"
};
#endif
#define FRTYPES 20	/* Total number of frame types in this array */

#if 0
static char *trailertypes[] = 
{
   "ZCRCE", /* 'h' CRC next, frame ends, header packet follows */
   "ZCRCG", /* 'i' CRC next, frame continues nonstop */
   "ZCRCQ", /* 'j' CRC next, frame continues, ZACK expected */
   "ZCRCW"  /* 'k' CRC next, ZACK expected, end of frame */
};
#endif


/* --| Sender |-------------------------------------------------------------- */

static guint8 
nibble_encode(int value)
{
  return value + (value<10?'0':('a'-10));
}

static void 
hex_encode(guint8* to, guint8* from, int len)
{ int i;
  for (i = 0; i < len; i++)
  { 
    to[2*i+0] = nibble_encode(from[i]/16);
    to[2*i+1] = nibble_encode(from[i]%16);
  }
}

static void 
chksum_encode(guint8* to, unsigned long value, int len)
{ int i;
  for(i = 0; i<len; i++, value /= 256)
    to[len-1-i] = value%256;
}

static void 
number_encode(guint8* to, unsigned long value, int len)
{ int i;
  for(i = 0; i<len; i++, value /= 256)
    to[i] = value%256;
}

void 
zmodem_snd_hex_hdr(guint8 typ, unsigned long info)
{ 
	guint8 hdr[7]; 
	guint8 buf[14];
	guint8 prefix[] = { ZPAD, ZPAD, ZDLE, ZHEX};   
	guint8 suffix[] = { '\r', '\n', 021}; 
	hdr[0] = typ;
	number_encode(hdr+1,info,4);
	chksum_encode(hdr+5,crc16((char*)hdr,5),2);
	hex_encode(buf,hdr,7);

	term->from_term(prefix, strlen(prefix));
	term->from_term(buf, 14);
	term->from_term(suffix, strlen(suffix));
}


#define combine(A,B) ((((guint16)(A))<<8)|(B))

void zmodem_receiver_rcv_frame(guint8 type, guint8 trailer, guint8* info, int len)
{
//printf("Frame [%10s %5s %4d]\r\n",
//     frametypes[type],
//     trailertypes[trailer-ZCRCE],
//     len
//    ); fflush(stdout);
//hexdump("",info,len);
  switch(combine(type,trailer))
  {
    case combine(ZCOMMAND,ZCRCW):
/*FIXME: improve */
//printf("Command: >%s<\r\n",info);
      zmodem_snd_hex_hdr(ZRPOS,0); /* or: ZSKIP */
      break;
    case combine(ZFILE,ZCRCW):
	filesize=0; tim=0; mode=0; nfilesleft=0; totalleft=0;
	sscanf((char*)info+strlen((char*)info)+1,"%lu %lo %o 0 %d %ld",
		&filesize, &tim, &mode, &nfilesleft, &totalleft);
//printf("Text: >%s<\r\n",(char*)info+strlen((char*)info)+1);
//printf("-------------- File -------------------- ---- Left ----\r\n");
//printf("Name---------- ----Size -----Time --Mode Files ---Bytes\r\n");
//printf("%14s %8ld %9ld %6o %5d %8ld\r\n",info,size,tim,mode,filesleft,totalleft);

      curr_file_offset = 0;
	  curr_file = new_file(info);
		if(curr_file == NULL) {
			zmodem_skip();
			break;
		}
      zmodem_snd_hex_hdr(ZRPOS,curr_file_offset); /* or: ZSKIP */
      break;
    case combine(ZDATA,ZCRCE): /*  frame ends, header packet follows */
      /*FIXME: eof processing*/
      fwrite(info,len,1,curr_file); /*FIXME: check fwrite successfull*/
      curr_file_offset += len;
      break;
    case combine(ZDATA,ZCRCG): /* frame continues nonstop */
      fwrite(info,len,1,curr_file); /*FIXME: check fwrite successfull*/
      curr_file_offset += len;
      break;
    case combine(ZDATA,ZCRCQ): /* frame continues, ZACK expected */
      fwrite(info,len,1,curr_file); /*FIXME: check fwrite successfull*/
      curr_file_offset += len;
      zmodem_snd_hex_hdr(ZACK,curr_file_offset); /* or: ZRPOS */
      break;
    case combine(ZDATA,ZCRCW): /* ZACK expected, end of frame */
      fwrite(info,len,1,curr_file); /*FIXME: check fwrite successfull*/
      curr_file_offset += len;
      zmodem_snd_hex_hdr(ZACK,curr_file_offset); /* or: ZRPOS */
      break;
    default:
      zmodem_receiver_rcv_error(); /* data (with this trailer) unexpected for type */
      break;
  }
}

void zmodem_receiver_rcv_header(guint8 type, guint32 info)
{
//printf("Header[%10s 0x%08lx]\r\n",frametypes[type],info); fflush(stdout);
	switch (type)
	{
	case ZRQINIT :
	  zmodem_snd_hex_hdr(ZRINIT,0);
	  break;
	case ZFILE   :
	  break;
	case ZCOMMAND:
	  break;
	case ZDATA   : /* FIXME: check ZFILE before */
	  /*FIXME: check for correct offset ... */
	  if (curr_file_offset != info)
		printf("----------> something wrong with sync\r\n");
	  break;
	case ZEOF    : /* FIXME: check completed ZDATA before (general check) */
	  fclose(curr_file); /*FIXME: check close successfull*/
	  /*FIXME: or ZFERR or ignore(more ZDATA???) */
	  zmodem_snd_hex_hdr(ZRINIT,0); 
	  zmodem_dmy_doneFile();
	  break;
	case ZFIN    : 
	  zmodem_dmy_doneTransmission();
	  zmodem_snd_hex_hdr(ZFIN,0);
	  /*FIXME: handle expect "OO"*/
	  zmodem_done();
	  break;
	}
}

void 
zmodem_receiver_rcv_error()
{
/*FIXME: recover*/
/*FIXME: timeout: 8192 bit (assuming 2400 baud) gives 5 sec */
/*must keep some kind of state*/
  zmodem_dmy_newError();
}


/* --| Zmodem Scanner |------------------------------------------------------ */

#define inRange(L,X,H) (((L) <= (X)) && (X) <= (H))
#define isNibble(X)    (inRange('0',(X),'9')||inRange('a',(X),'f'))
#define isHeader(C)    (inRange('A',(C),'D'))
#define isTrailer(C)   (inRange('h',(C),'k'))

static int 
nibble_decode(guint8 nibble)
{
	return nibble - (nibble < '9' ? '0' : ('a'-10));
}

static void 
hex_decode(guint8* to, guint8* from, int len)
{ 
	int i;
	for (i = 0; i < len; i++)
		to[i] = 16*nibble_decode(from[2*i+0])+nibble_decode(from[2*i+1]);
}

#if 0
static unsigned long 
chksum_decode(guint8* buf, int len)
{ 
	int i; 
	unsigned long res = 0;
	for (i=0; i < len; i++) res = 256*res+buf[i];
	return res;
}
#endif

static unsigned long 
number_decode(guint8* buf, int len)
{ 
	int i; 
	unsigned long res = 0;
	for (i=len-1; i >= 0; i--) res = 256*res+buf[i];
	return res;
}

static int 
zdle_decode(guint8* to, guint8* from, int len)
{ 
	int i,j;
  for (i = 0, j=0; i < len; i++)
  {
    if (from[i] == ZDLE)
    {
      if (i+1 > len)                  return -1; /* fault */
      if ((from[i+1] & 0x60) != 0x40) return -2; /* fault */
      /* FIXME: special treatment for 0x177, 0x3ff */
      to[j++] = from[++i]^0x40;
    }
    else
      to[j++] = from[i];
  }
  return j;
}

void 
zmodem_rcv_error(char* comment, guint8* buffer, int len)
/* received error */
{
	hexdump(comment,(char*)buffer,len);
	zmodem_receiver_rcv_error();
}

/* abort zmodem */
void 
zmodem_done(/*FIXME: bool ok*/)
{
	end_internal_rz();
}

/* abort zmodem */
void 
zmodem_rcv_fatal(char* message)
{
	fprintf(stderr,"fatal error: %s\n",message);
	zmodem_done();
}

/* received a header in buffer */
void 
zmodem_rcv_header()
{ 
	int i; 
	guint8 hdr[5+4];
	last_hdr_typ=-1;
  switch(buffer[1])
  {
    case ZBIN :
      if( zdle_decode(hdr,buffer+2,buftop-2) != 7)
      { zmodem_rcv_error("BAD BIN HEADER (zdle)",buffer,buftop); return; }
      if (crc16((char*)hdr,7) != 0)
      { zmodem_rcv_error("BAD BIN HEADER (chksum)",buffer,buftop); return; }
      break;
    case ZHEX :
      /* buffer[2..15] contains nibbles, buffer[16..17] = CR LF */
      for (i = 2; i <= 15; i++) if (!isNibble(buffer[i]))
      { zmodem_rcv_error("BAD HEX HEADER (nibble)",buffer,buftop); return; }
      if ((buffer[16]&0x7f) != '\r' || (buffer[17]&0x7f) != '\n')
      { zmodem_rcv_error("BAD HEX HEADER (cr lf)",buffer,buftop); return; }
      hex_decode(hdr,buffer+2,7);
      if (crc16((char*)hdr,7) != 0)
      { zmodem_rcv_error("BAD HEX HEADER (chksum)",buffer,buftop); return; }
      break;
    default:
      zmodem_rcv_fatal("unknown header type");
      break;
  }
  /* last check before prime time */
  if (hdr[0] > FRTYPES)
  { char buf[100]; sprintf(buf,"Invalid buffer type %d(%c)",hdr[0],hdr[0]);
    zmodem_rcv_fatal(buf);
    return;
  }
  /* real header processing */
  last_hdr_typ = hdr[0];
  zmodem_receiver_rcv_header(hdr[0],number_decode(hdr+1,4));
}

void zmodem_rcv_frame()
/* received a frame in buffer */
{ int len; guint8 frame[3000];
  len = zdle_decode(frame,frame_buffer,frame_buftop);
  if (len < 0) 
  { zmodem_rcv_error("Bad Frame (zdle)",frame_buffer,frame_buftop); return; }
  frame[len] = buffer[1];
  if (zdle_decode(frame+len+1,buffer+2,buftop-2) != 2)
  { zmodem_rcv_error("Bad Trailer (zdle)",buffer,buftop); return; }
  if (crc16((char*)frame,len+3) != 0)
  { zmodem_rcv_error("Bad Frame (chksum)",buffer,buftop); return; }
  zmodem_receiver_rcv_frame(last_hdr_typ, buffer[1], frame, len);
  if (buffer[1] == ZCRCW || buffer[1] == ZCRCQ) last_hdr_typ = -1;
}

/* received cancel (CAN*5) */
void 
zmodem_rcv_cancel()
{
  hexdump("CANCEL",NULL,0);
  zmodem_done();
}


static int hdr_len[4] = { 9,18 }; /* header length modulo ZDLE codes */

/* --| low level packet scanner |-------------------------------------------- */

/*
   What follows, is the workhorse of this module. It splits all incoming
   guint8 up into headers and frames while doing the nessesary decoding
   and crc checking.
*/

//static char* zmodem_keyword = "rz\r**";
//static char* zmodem_keyword = "rz\r**\030B000000000000\r\n";


/* process a guint8 just received */
/*FIXME: better use signals, MUST be void function */
/* OK, this function looks like hell, but it isn't */
/* We simply buffer characters here, and make see  */
/* then, if we can do something with the buffer    */
/*
  A | * ZDLE ZDLE ZDLE ZDLE ZDLE                  | cancel
  B | * ZDLE <Header>                             | initial stuff is dump
    | * ZDLE <Trailer>                            | initial stuff is packet
  C |   ZDLE ZBIN16  T x x x x CrC CrC            | 16 bit crc header
    |   ZDLE ZHEX    TT xx xx xx xx CrC CrC CR LF | hex header
  D |   ZDLE <Trailer>  CrC CrC                   | 16 bit data packet (crc16)
*/
void 
zmodem_rcv_byte(int cc)
{
  /* --| Scan for zmodem keyword |----------------------------------------- */


  /*--| Zmodem scanning starts here |------------------------------------- */

  if ((cc & 0175) == 021) return; /* zmodem ignores XON/XOFF */

/*FIXME: Break character? */

  if (buftop >= MAXBUFLEN)
  /* buffer overflow */
  {
    zmodem_rcv_error("Buffer overflow",buffer,buftop);
    buftop = 0;
  }
  buffer[buftop] = (guint8)cc;
  buftop        += 1;
  zdlecnt       += ((cc & 0xff) == ZDLE)?1:0;

  /* --| here come the real thing |---------------------------------------- */

  /* Check for patterns */

  /* A: CAN*5 cances sequence */
  if (buftop >= 5)
  { int i;
    for (i = 0; i < 5; i++) if (buffer[buftop-5+i] != ZDLE) goto nocan;
    /* if (buftop > 5) rcv_error("cancel anyway",buffer,buftop-5); */
    zmodem_rcv_cancel();
    buftop = 0; 
    zdlecnt = 0; 
    return;
  }
nocan:

  /* B1: header start */
  if (buftop >= 2 && buffer[buftop-2] == ZDLE && isHeader(buffer[buftop-1]))
  { int pos = buftop - 2;
    /* ignore possible ZPADs first */
    if (pos > 0 && buffer[pos-1] == ZPAD) pos -= 1;
    if (pos > 0 && buffer[pos-1] == ZPAD && buffer[buftop-1] == ZHEX) pos -= 1;
    if (pos > 0) zmodem_rcv_error("noise (hd follows)",buffer,pos);
    buffer[0] = ZDLE;
    buffer[1] = buffer[buftop-1];
    buftop = 2;
    zdlecnt = 1;
    return;
  }
  /* B1: frame start */
  if (buftop >= 2 && buffer[buftop-2] == ZDLE && isTrailer(buffer[buftop-1]))
  { int i;
    for (i = 0; i < buftop-2; i++) frame_buffer[i] = buffer[i];
    frame_buftop = buftop - 2;
    buffer[0] = ZDLE;
    buffer[1] = buffer[buftop-1];
    buftop = 2;
    zdlecnt = 1;
    return;
  }

  /* C: header end */
  if ( buffer[0] == ZDLE && isHeader(buffer[1]) && 
       buftop == hdr_len[buffer[1]-ZBIN]+zdlecnt-1)
  {
    zmodem_rcv_header();
    buftop = 0;
    zdlecnt = 0;
    return;
  }

  /* D: Frame end */
  if ( buftop == 4+zdlecnt-1 && buffer[0] == ZDLE && isTrailer(buffer[1]) )
  {
    zmodem_rcv_frame();
    buftop = 0;
    zdlecnt = 0;
  }     
  return;
}

/* --| stupidity functions |------------------------------------------------- */

void zmodem_dmy_newFile(const char* name, long size)
{
  //emit newFile(name, size);
}

void zmodem_dmy_newBlock(long size)
{
  //emit newBlock(size);
}

void zmodem_dmy_newError()
{
  //emit newError();
}

void zmodem_dmy_doneFile()
{
  //emit doneFile();
}

void zmodem_dmy_doneTransmission()
{
  //emit doneTransmission();
}


/* --| constructor |--------------------------------------------------------- */

void
zmodem_init(void)
{
	key_pos = 0;
	buftop = 0;
	zdlecnt = 0;
	last_hdr_typ=-1;
	zmodem_snd_hex_hdr(ZRINIT,0); /* here we are! */
}

void
zmodem_input_handler(gpointer data, int source , GdkInputCondition con)
{
	char buf[256];
	int i;
	int size = read(source, buf, 256);
	for(i=0;i<size;i++) zmodem_rcv_byte(buf[i]);
}

void
zmodem_cancel(void)
{
	static char canistr[] = {
	 24,24,24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8,8,8,0
	};
	term->from_term(canistr, 20);
}

void
zmodem_skip(void)
{
	zmodem_snd_hex_hdr(ZSKIP, 0);
}

FILE *
new_file(char *filename)
{
	char buf[100];
	struct stat st;
	char *s = filename;
	while(*s) {
		*s = tolower(*s);
		s++;
	}
	sprintf(buf,"%s/%s",expand_path(DOWN_PATH),filename);
	if(stat(buf, &st) == 0) {
		// success: this file exists already 
		// we will append new junks to the old file.
		bytes_skipped = st.st_size;
		curr_file = fopen(buf, "a");
	} else {
		curr_file = fopen(buf, "w");
	}
	if(curr_file == NULL) g_warning("fail to open %s for writing\n", buf); 
	return curr_file;
}
