
/* this file is a part of gau software, (C) Hwang chi-duck 1997               */

#include <gtk/gtk.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
//#include "config.h"
#include "gau.h"

enum { 
	WAITFOR, 
	WAITFOR_MULTI,
	TRANSMIT, 
	SETTITLE, 
	BOTTOMMENU,
	SLEEP,
	BEEP,
	GOTO,
	LABEL,
	PLAY,
	CAPTURE_ON,
	CAPTURE_OFF,
	DISCONNECT,
	INCLUDE,
	END};

typedef struct _Script Script;
struct _Script {
	int type;
	gpointer *arg;
};

static GList *current_script_item = NULL;
static GList  *telnet_script = NULL;

static char *wait_string = NULL;
static int waitfor_timeout_tag = -1;
static gboolean waitfor_waiting = FALSE;

static int idle_tag = -1;
static int sleep_tag = -1;
static gboolean script_capture_active = FALSE;

static GList * make_script_from_file(char *filename);
static GList * make_script_from_string(char *str);
static GList * make_script_from_command(char *command);

static void end_script(void);
static int waitfor_filter(char *buf, int size);
static int waitfor_timeout(gpointer data);
static int next_script(gpointer);
static void script_set_title(char *title);
static void script_bottom_menu_update(char *menu_file);
static void run_telnet_login_script(void);

void
install_next_script(void)
{
	gau_state = SCRIPT;
	idle_tag = gtk_timeout_add(0, next_script, NULL);
}

void
waitfor(char *buf, int second)
{
	char tmp_buf[256];

	input_filter_list = g_list_append(input_filter_list, waitfor_filter);

	if(wait_string) g_free(wait_string);
	wait_string = g_strdup(buf); 
	waitfor_filter(wait_string, 0); // initialize
	waitfor_waiting = TRUE;

	sprintf(tmp_buf, "\"%s\" ٸϴ", wait_string);
	show_message(tmp_buf);

	if(second > 0)
		waitfor_timeout_tag = gtk_timeout_add(second, waitfor_timeout, NULL);
}

int
waitfor_timeout(gpointer data)
{
	waitfor_waiting = FALSE;
	input_filter_list = g_list_remove(input_filter_list, waitfor_filter);
	if(current_script_item->next == NULL) {
		end_script();
	} else {
		current_script_item = current_script_item->next;
		install_next_script();
	}
	waitfor_timeout_tag = -1;
	return FALSE;
}

int
waitfor_filter(char *buf, int size)
{
	int i;
	static char *wait_string_p;
	if(size == 0) wait_string_p = wait_string;
	for(i=0;i<size;i++) {
		if(buf[i] != *wait_string_p) wait_string_p = wait_string;
		else if(*++wait_string_p == '\0') {
			input_filter_list = 
				g_list_remove(input_filter_list, waitfor_filter);
			waitfor_waiting = FALSE;
			if(waitfor_timeout_tag != -1)
				gtk_timeout_remove(waitfor_timeout_tag);
			waitfor_timeout_tag = -1;
			if(current_script_item->next == NULL) {
				destroy_script();
			} else {
				current_script_item = current_script_item->next;
				install_next_script();
			}
		}
	}
	return FALSE;
}

int
waitformult_filter(char *buf, int byte)
{
	return FALSE;
}

void
waitformulti(void)
{
}

void
transmit(char *str)
{
	term->from_term(str, strlen(str));
}

void
script_goto(char *str)
{
	GList *list;
	Script *script;
	list = current_script_item;
	while(list) {
		script = list->data;
		if(script->type == LABEL) {
			if(strcmp(script->arg[0], str) == 0) {
				current_script_item = list;
				return;
			}
		}
		list = list->prev;
	}
	list = current_script_item->next;
	while(list) {
		script = list->data;
		if(script->type == LABEL) {
			if(strcmp(script->arg[0], str) == 0) {
				current_script_item = list;
				return;
			}
		}
		list = list->next;
	}
	g_warning("invalid goto: %s\n", str);
}

void
script_sleep(int msec)
{
	sleep_tag = gtk_timeout_add(msec, next_script, NULL);
	current_script_item = current_script_item->next;
}


void
run_login_script(void)
{
	extern PhoneItem *current_item;
	if(!use_modem) {
		run_telnet_login_script();
		return;
	}

	if(current_item == NULL || current_item->script_file == NULL) return;
	sprintf(tmpbuf, "%s/%s", expand_path(script_path), current_item->script_file);
	run_script_from_file(tmpbuf);
}
	

static char *
get_next_token(char *buf, char *d)
{
	char *s = buf;
	while(isspace(*s)) s++;
	if(*s == '"') {
		*d++ = *s++;
		while(*s != '"') *d++ = *s++;
		*d++ = *s++;
	} else {
		while(!isspace(*s)) *d++ = *s++;
	}
	*d = 0;
	return s;
}

static int
mextract_string(char *src, char *des)
{
	if(*src++ != '"') return FALSE;
	while(*src != '"') {
		if(*src == '\\') {
			src++;
			switch(*src) {
				case 'r':
					*des++ = '\r';
					break;
				case 'n':
					*des++ = '\n';
					break;
				default:
					*des++ = *src;
					break;
			}
			src++;
		} else if(*src == '^') {
			src++;
			*des++ = *src++ + '\b' - 'H';
		} else {
			*des++  = *src++;
		}
	}
	*des = 0;
	return TRUE;
}

static Script *
parse_line(char *buf)
{
	Script *script = NULL;
	char *s;
	char tmp[250], tmp2[250];

	s = get_next_token(buf, tmp);
	if(strcmp(tmp, "waitfor") == 0) {
		get_next_token(s, tmp);
		mextract_string(tmp, tmp2);
		script = g_new(Script, 1);
		script->type = WAITFOR;
		script->arg = g_new(gpointer, 2);
		script->arg[0] = g_strdup(tmp2);
		script->arg[1] = 0;
	} else if(strcmp("transmit", tmp) == 0) {
		get_next_token(s, tmp);
		mextract_string(tmp, tmp2);
		script = g_new(Script, 1);
		script->type = TRANSMIT;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = g_strdup(tmp2);
	} else if(strcmp("set-title", tmp) == 0) {
		get_next_token(s, tmp);
		mextract_string(tmp, tmp2);
		script = g_new(Script, 1);
		script->type = SETTITLE;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = g_strdup(tmp2);
	} else if(strcmp("bottom-menu", tmp) == 0) {
		get_next_token(s, tmp);
		mextract_string(tmp, tmp2);
		script = g_new(Script, 1);
		script->type = BOTTOMMENU;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = g_strdup(tmp2);
	} else if(strcmp("play", tmp) == 0) {
		get_next_token(s, tmp);
		mextract_string(tmp, tmp2);
		script = g_new(Script, 1);
		script->type = PLAY;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = g_strdup(tmp2);
	} else if(strcmp("beep", tmp) == 0) {
		script = g_new(Script, 1);
		script->type = BEEP;
	} else if(strcmp("capture", tmp) == 0) {
		s = get_next_token(s, tmp);
		script = g_new(Script, 1);
		if(strcmp(tmp, "on") == 0) {
			get_next_token(s, tmp);
			mextract_string(tmp, tmp2);
			script->type = CAPTURE_ON;
			script->arg = g_new(gpointer, 1);
			script->arg[0] = g_strdup(tmp2);
		} else if(strcmp(tmp, "off") == 0) {
			script = g_new(Script, 1);
			script->type = CAPTURE_OFF;
		}
	} else if(strcmp("end", tmp) == 0) {
		return NULL;
	} else if(strcmp("disconnect", tmp) == 0) {
		script = g_new(Script, 1);
		script->type = DISCONNECT;
	} else if(strcmp("sleep", tmp) == 0) {
		get_next_token(s, tmp);
		script = g_new(Script, 1);
		script->type = SLEEP;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = (gpointer)((int)(1000.0*atof(tmp)));
	} else if(strcmp("goto", tmp) == 0) {
		get_next_token(s, tmp);
		script = g_new(Script, 1);
		script->type = GOTO;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = g_strdup(tmp);
	} else if(strcmp("include", tmp) == 0) {
		get_next_token(s, tmp);
		mextract_string(tmp, tmp2);
		script = g_new(Script, 1);
		script->type = INCLUDE;
		script->arg = g_new(gpointer, 1);
		script->arg[0] = g_strdup(tmp2);
		return script;
	} else {
		s = tmp;
		while(*s && !isspace(*s)) s++;
		s--;
		if(*s == ':') {
			*s = 0;
			script = g_new(Script, 1);
			script->type = LABEL;
			script->arg = g_new(gpointer, 1);
			script->arg[0] = g_strdup(tmp);
		} else 
			g_warning("unknown script \n");
	}
	return script;
}

static GList *
make_script_from_file(char *filename)
{
	FILE *fp;
	Script *script;
	GList *list;
	GList *sublist, *last;
	char *buf;
	if(filename == NULL || filename[0] == 0) {
		g_warning("filename is empty in %s\n", __FUNCTION__);
		return NULL;
	}
	fp = fopen(filename, "r");
	if(fp == NULL) {
		g_warning("fail to open %s\n", filename);
		return NULL;
	}

	list = NULL;
	while((buf = get_line(fp)) != NULL) {
		script = parse_line(buf);
		if(script == NULL) break;
		if(script->type == INCLUDE) {
			sprintf(tmpbuf, "%s/%s", expand_path(script_path), (char*)script->arg[0]);
			sublist = make_script_from_file(tmpbuf);
			if(sublist != NULL) {
				if(list == NULL) list = sublist;
				else {
					last = g_list_last(list);
					last->next = sublist;
					sublist->prev = last;
				}
			}
			g_free(script->arg[0]);
			g_free(script->arg);
			g_free(script);
		} else
			list = g_list_append(list, script);
	}
	fclose(fp);
	return list;
}

static GList *
make_script_from_string(char *str)
{
	GList *list;
	Script *script;
	char *s;

	list = NULL;

	s = strtok(str, ";");
	while(s != NULL) {
		script = parse_line(s);
		if(script == NULL) break;
		list = g_list_append(list, script);
		s = strtok(NULL, ";");
	}
	return list;
}

static GList *
make_script_from_command(char *command)
{
	FILE *fp;
	GList *list;
	Script *script;
	char *buf;
	void (*chld_handler)(int);
	chld_handler = signal(SIGCHLD, SIG_DFL);
	fp = popen(command, "rw");
	if(fp == NULL) {
		g_warning("fail to run %s\n", command);
		return NULL;
	}
	list = NULL;
	while((buf = get_line(fp)) != NULL) {
		script = parse_line(buf);
		if(script == NULL) break;
		list = g_list_append(list, script);
	}
	pclose(fp);
	signal(SIGCHLD, chld_handler);
	return list;
}

void
destroy_script(void)
{
	Script *script;
	GList *first, *list;
	first = g_list_first(current_script_item);
	list = first;
	while(list) {
		script = list->data;
		switch(script->type) {
			case WAITFOR: case BOTTOMMENU: case SETTITLE: 
			case CAPTURE_ON: case PLAY: case LABEL: case GOTO:
				g_free(script->arg[0]);
				g_free(script->arg);
				break;
			case WAITFOR_MULTI:
				g_free(script->arg);
				break;
			case SLEEP:
				g_free(script->arg);
				break;
			default:
				break;
		}
		g_free(script);
		list = list->next;
	}
	g_list_free(first);
}


void
script_cancel(void)
{
	if(waitfor_waiting) {
		if(waitfor_timeout_tag > 0) gtk_timeout_remove(waitfor_timeout_tag);
		input_filter_list = g_list_remove(input_filter_list, 
			waitfor_filter);
		waitfor_waiting =  FALSE;
	}
	if(idle_tag != -1) gtk_timeout_remove(idle_tag);
	if(sleep_tag != -1) gtk_timeout_remove(sleep_tag);
	destroy_script();
	if(script_capture_active) end_capture();
	current_script_item = NULL;
}


int
next_script(gpointer foo)
{
	Script *script;
	int type = -1;
	idle_tag = -1;
	sleep_tag = -1;
	if(current_script_item == NULL) {
		g_warning("script is null\n");
		return FALSE;
	} 

	script = current_script_item->data;
	switch(type = script->type) {
		case WAITFOR:
			waitfor(script->arg[0], (int)script->arg[1]);
			break;
		case TRANSMIT:
			transmit(script->arg[0]);
			break;
		case SETTITLE:
			script_set_title(script->arg[0]);
			break;
		case BOTTOMMENU:
			script_bottom_menu_update(script->arg[0]);
			break;
		case BEEP:
			gdk_beep();
			break;
		case PLAY:
			play(script->arg[0]);
			break;
		case DISCONNECT:
			hangup();
			break;
		case CAPTURE_ON:
			script_capture_active = TRUE;
			begin_capture(script->arg[0]);
			break;
		case CAPTURE_OFF:
			script_capture_active = FALSE;
			end_capture();
			break;
		case SLEEP:
			script_sleep((int)script->arg[0]);
			break;
		case GOTO:
			script_goto(script->arg[0]);
			break;
		case LABEL:
			break;
		default:
			g_print("%p script_list\n", current_script_item);
			g_warning("unknown type : in next-script");
			gau_state = NOTHING;
			current_script_item = NULL;
			return FALSE;
			break;
	}
	if(current_script_item->next == NULL && script->type != WAITFOR && script->type != WAITFOR_MULTI && script->type != SLEEP) {
		end_script();
		return FALSE;
	}

	if(script->type != WAITFOR  && 
		script->type != WAITFOR_MULTI &&
		script->type != SLEEP && 
		current_script_item->next) {
		current_script_item = current_script_item->next;
		script = current_script_item->data;
		if(script->type == WAITFOR || script->type == WAITFOR_MULTI) {
			waitfor(script->arg[0], (int)script->arg[1]);
		} else
			install_next_script();
	}
	return FALSE;
}

static void
end_script(void)
{
	destroy_script();
	current_script_item = NULL;
	show_message("ȥڸ ϴ");
	gau_state = NOTHING;
}

int
delayed_timeout(char *buf)
{
	term->from_term(buf, strlen(buf));
	g_free(buf);
	return FALSE;
}

void
delayed_transmit(char *s, int bytes, int msec)
{
	char *buf = g_new(char, bytes+1);
	int i;
	for(i=0;i<bytes;i++) buf[i] = s[i];
	buf[i] = 0;
	gtk_timeout_add(msec, (GtkFunction)delayed_timeout, buf);
}

void
prepare_telnet_script(char *script_file)
{
	telnet_script = NULL;
	if(script_file == NULL || script_file[0] == 0) return;
	sprintf(tmpbuf, "%s/%s", expand_path(script_path), script_file);
	telnet_script = make_script_from_file(tmpbuf);
}

static void
run_telnet_login_script(void)
{
	if(telnetinfo && telnet_script) {
		current_script_item = telnet_script;
		install_next_script();
	}
}

void
run_script_from_file(char *filename)
{
	GList *list;
	list = make_script_from_file(filename);
	if(list == NULL) {
		g_warning("fail to run %s\n", filename);
		return;
	}
	current_script_item = list;
	install_next_script();
}

void
run_script_from_string(char *str)
{
	GList *list;
	list = make_script_from_string(str);
	if(list == NULL) {
		g_warning("fail to run %s ...\n", str);
		return;
	}
	current_script_item = list;
	install_next_script();
}

void
run_script_from_command(char *command)
{
	GList *list;
	list = make_script_from_command(command);
	if(list == NULL) {
		g_warning("fail to extract from command %s\n", command);
		return;
	}
	current_script_item = list;
	install_next_script();
}

void
script_set_title(char *title)
{
	strcpy(tmpbuf, ": ");
	strcat(tmpbuf, title);
	gtk_window_set_title(GTK_WINDOW(main_win), tmpbuf);
}

void
script_bottom_menu_update(char *menu_file)
{
	update_bottom_menu_from_file(menu_file);
}
