#include <unistd.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <string.h>


#include "gau.h"
#include "config.h"
#include <gdk/gdkkeysyms.h>

int srz_pid = -1;
int use_internal_rz = TRUE;
gboolean exit_after_download = FALSE;
GtkWidget *sz_filename_entry;
GtkWidget *sz_query_win;
static void run_real_sz(char *filename);
static int grab_handle_event (GtkWidget *widget, GdkEvent *event);

int
zmodem_input_filter(char *str, int bytes)
{
	static char           zmSig[] = "**\030B0";
	static char          *zmSigPtr = zmSig;
	int i;
	/* check signature of zmodem */
	for(i=0;i<bytes;i++) {
		if(*zmSigPtr == '\0') {
			if(str[i] == '0') {
				delayed_run(100, run_rz);
				zmSigPtr = zmSig;
				return FALSE;
			} else if(str[i] == '1') {
				delayed_run(100, run_sz);
				zmSigPtr = zmSig;
				return FALSE;
			}
			zmSigPtr = zmSig;
		} else {
			if(str[i] != *zmSigPtr) zmSigPtr = zmSig;
			else zmSigPtr++;
		}
	}
	return FALSE;
}

void rz_signal_child_handle(int sig);
void sz_signal_child_handle(int sig);

int srz_input_tag;
int srzpipefd[2];

void 
srz_input_handler(gpointer data, int source, GdkInputCondition con)
{
	char buf[256];
	int size = read(source, buf, 255);
	term->to_term(term, buf, size);
}


void
run_rz(void)
{
	gau_state = DOWNLOAD;
	gtk_widget_set_sensitive(hangup_after_transfer_button, TRUE);
	if(use_internal_rz) {
		run_internal_rz();
		return;
	}
	pipe(srzpipefd);
	srz_pid = fork();
	if(srz_pid > 0) {
		/* parent */
		signal(SIGCHLD, rz_signal_child_handle);
		gdk_input_remove(read_tag);
		close(srzpipefd[1]);
		srz_input_tag = 
			gdk_input_add(srzpipefd[0], GDK_INPUT_READ, 
				srz_input_handler, NULL);
		return;
	}
	chdir(expand_path(DOWN_PATH));
	dup2(read_fd, 0);
	dup2(read_fd, 1);
	close(read_fd);
	close(srzpipefd[0]);
	dup2(srzpipefd[1], 2);
	execlp("lrz", "lrz", "-r", "-vv", NULL);
	g_warning("%s : fail to run rz\n", __FUNCTION__);
	_exit(1);
}

static void
sz_send(void)
{
	char *s = gtk_entry_get_text(GTK_ENTRY(sz_filename_entry));
	run_real_sz(s);
	gtk_widget_destroy(sz_query_win);
}

static void
sz_dialog_key_press(GtkWidget *w, GdkEventKey *ev, gpointer data)
{
	if(ev->keyval == GDK_Return) {
		sz_send();
	}
}

void
sz_initial_cancel(GtkWidget *w, GtkWidget *win)
{
	static char canistr[] = {
	 24,24,24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8,8,8,0
	};
	term->from_term(canistr, strlen(canistr));
	if(win) gtk_widget_destroy(win);
}

void
sz_file_selection_ok(GtkWidget *w, GtkWidget *win)
{
	char *filename = gtk_file_selection_get_filename(GTK_FILE_SELECTION(win));
	if(getcwd(tmpbuf, 256) != NULL) {
		int n = strlen(tmpbuf);
		if(strncmp(filename, tmpbuf, n) == 0 && filename[n] == '/') {
			gtk_entry_set_text(GTK_ENTRY(sz_filename_entry), filename + n + 1);
		} else {
			gtk_entry_set_text(GTK_ENTRY(sz_filename_entry), filename);
		}
	} else {
		gtk_entry_set_text(GTK_ENTRY(sz_filename_entry), filename);
	}
	gtk_widget_destroy(win);
	gtk_grab_add(sz_query_win);
}

void
sz_file_selection_cancel(GtkWidget *w, GtkWidget *win)
{
	gtk_widget_destroy(win);
	gtk_grab_add(sz_query_win);
}

void
sz_file_search(GtkWidget *w, gpointer data)
{
	GtkWidget *file_selection = gtk_file_selection_new("  ãƺ");
	gtk_window_position(GTK_WINDOW(file_selection), GTK_WIN_POS_MOUSE);
	gtk_signal_connect(
		GTK_OBJECT(GTK_FILE_SELECTION(file_selection)->ok_button),
		"clicked",
		GTK_SIGNAL_FUNC(sz_file_selection_ok), 
		file_selection);
	gtk_signal_connect(
		GTK_OBJECT(GTK_FILE_SELECTION(file_selection)->cancel_button),
		"clicked",
		GTK_SIGNAL_FUNC(sz_file_selection_cancel), 
		file_selection);
	gtk_widget_destroy(GTK_FILE_SELECTION(file_selection)->help_button);
	gtk_grab_remove(sz_query_win);
	gtk_widget_show(file_selection);
}

void
run_sz(void)
{
	GtkWidget *vbox, *button, *label, *hbox;
	gtk_widget_set_sensitive(hangup_after_transfer_button, TRUE);
	sz_query_win = gtk_window_new(GTK_WINDOW_DIALOG);
	gtk_container_border_width(GTK_CONTAINER(sz_query_win), 10);
	gtk_window_set_title(GTK_WINDOW(sz_query_win), " ");
	gtk_window_position(GTK_WINDOW(sz_query_win), GTK_WIN_POS_MOUSE);
	vbox = gtk_vbox_new(FALSE, 2);
	gtk_container_add(GTK_CONTAINER(sz_query_win), vbox);
	gtk_widget_show(vbox);

	hbox = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_widget_show(hbox);

	label = gtk_label_new(" ̸:");
	gtk_container_add(GTK_CONTAINER(hbox), label);
	gtk_widget_show(label);

	sz_filename_entry = gtk_entry_new();
	gtk_container_add(GTK_CONTAINER(hbox), sz_filename_entry);
	gtk_signal_connect(GTK_OBJECT(sz_filename_entry), "key_press_event",
		(GtkSignalFunc)sz_dialog_key_press, sz_query_win);
	gtk_widget_show(sz_filename_entry);
	gtk_widget_new(gtk_hseparator_get_type(),
		"GtkWidget::parent", vbox,
		"GtkWidget::visible", TRUE,
		NULL);
	hbox = gtk_hbox_new(FALSE, 10);
	gtk_container_add(GTK_CONTAINER(vbox), hbox);
	gtk_widget_show(hbox);

	button = gtk_button_new_with_label("ãƺ");
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)sz_file_search, sz_query_win);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("");
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)sz_initial_cancel, sz_query_win);
	gtk_widget_show(button);

	button = gtk_button_new_with_label("");
	gtk_container_add(GTK_CONTAINER(hbox), button);
	gtk_signal_connect(GTK_OBJECT(button), "clicked",
		(GtkSignalFunc)sz_send, sz_query_win);
	gtk_widget_show(button);

	gtk_widget_show(sz_query_win);
	gtk_widget_grab_focus(sz_filename_entry);
	gtk_signal_connect(GTK_OBJECT(sz_query_win), "event",
		GTK_SIGNAL_FUNC(grab_handle_event), NULL);
	gtk_grab_add(sz_query_win);
}


static void
run_real_sz(char *filename)
{
	struct stat buf;
	if(filename == NULL) {
		g_warning("In sz, filename is NULL!!!\n");
		return;
	}
	if(stat(filename, &buf) != 0) {
		g_warning("In \"%s\", fail to read '%s'\n", __FUNCTION__, filename);
		sz_initial_cancel(NULL, NULL);
		return;
	}
	if(!use_modem && use_ztelnet) {
		sprintf(tmpbuf, "%c", ']' & 0x1f);
		term->from_term(tmpbuf, strlen(tmpbuf));
		usleep(20000);
		sprintf(tmpbuf, "%s %s\r", sz_command, filename);
		term->from_term(tmpbuf, strlen(tmpbuf));
		return;
	}

	pipe(srzpipefd);
	srz_pid = fork();
	if(srz_pid > 0) {
		/* parent */
		signal(SIGCHLD, sz_signal_child_handle);
		gdk_input_remove(read_tag);
		close(srzpipefd[1]);
		srz_input_tag = gdk_input_add(srzpipefd[0], GDK_INPUT_READ, srz_input_handler, NULL);
		gau_state = UPLOAD;
		return;
	}
	// child
	dup2(read_fd, 0);
	dup2(read_fd, 1);
	close(read_fd);
	dup2(srzpipefd[1], 2);
	close(srzpipefd[0]);
	sprintf(tmpbuf, "%s %s", sz_command, filename);
	execlp("sh", "sh", "-c", tmpbuf, NULL);
	g_warning("%s : fail to run sz\n", __FUNCTION__);
	_exit(1);
}

int
confirm_death_of_rz(void)
{
	if(srz_pid > 0) {
		// still alive.
		kill(srz_pid, SIGTERM);
	}
	return FALSE; // just once
}


void
rz_cancel(void)
{
	static char canistr[] = {
	 24,24,24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8,8,8,0
	};

	if(srz_pid <= 0) return;
	// first ask remote side to quit
	// ̹  sz  kill ؾ.
	term->from_term(canistr, strlen(canistr));
	gtk_timeout_add(1000, (GtkFunction)confirm_death_of_rz, NULL);
}

void
sz_cancel(void)
{
	if(srz_pid <= 0) return;
	kill(srz_pid, SIGTERM);
}

void
zmodem_signal_restore(void)
{
	if(use_modem) signal(SIGCHLD, SIG_IGN);
	else if(!use_internal_telnet) {
		signal(SIGCHLD, telnet_child_signal);
	} else { // using internal telnet
		signal(SIGCHLD, SIG_IGN);
	}
}

void
rz_signal_child_handle(int sig)
{
	int status;
	if(wait(&status) != srz_pid) {
		signal(SIGCHLD, rz_signal_child_handle);
		return;
	}
	if(!WIFEXITED(status) || (WIFEXITED(status) && WEXITSTATUS(status) != 0)) {
		// killed by a signal which was not caught;
		// or exit code is not zero;
		static char canistr[] = {
		 24,24,24,24,24,24,24,24,24,24,8,8,8,8,8,8,8,8,8,8,0
		};
		g_print("rz exit abnormally\n");
		term->from_term(canistr, strlen(canistr));
	} 
	zmodem_signal_restore();
	if(exit_after_download) {
		hangup();
	} 
	gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
	play_transfer_end();
	srz_pid = -1;
	gdk_input_remove(srz_input_tag);
	close(srzpipefd[0]);
	read_tag = gdk_input_add(read_fd, GDK_INPUT_READ, 
		input_handler_from_external, NULL);
}

void
sz_signal_child_handle(int sig)
{
	if(wait(NULL) != srz_pid) {
		signal(SIGCHLD, sz_signal_child_handle);
		return;
	}
	gdk_input_remove(srz_input_tag);
	read_tag = gdk_input_add(read_fd, GDK_INPUT_READ, 
		input_handler_from_external, NULL);
	zmodem_signal_restore();
	if(exit_after_download) {
		hangup();
	} else 
	gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
	play_transfer_end();
	srz_pid = -1;
}

int zmodem_input_tag;

void
run_internal_rz(void)
{
	zmodem_init();
	gdk_input_remove(read_tag);
	zmodem_input_tag = gdk_input_add(read_fd, 
		GDK_INPUT_READ, zmodem_input_handler, NULL);
}

void
end_internal_rz(void)
{
	gdk_input_remove(zmodem_input_tag);
	read_tag = gdk_input_add(read_fd, GDK_INPUT_READ, 
		input_handler_from_external, NULL);
	play_transfer_end();
	gtk_widget_set_sensitive(hangup_after_transfer_button, FALSE);
}


/* Events sent to the grab window could be anything - check if they
   correspond to a child of the grab window, and if so forward them.
   If they aren't destined for a child, or for the grab window, throw
   them out */
int
grab_handle_event (GtkWidget *widget, GdkEvent *event)
{
  GtkWidget *event_widget;
  GtkWidget *w;

  /* Find the window in which the event occurred */
  gdk_window_get_user_data (event->any.window, (void**) &event_widget);

  /* If it was really for the grab window, continue processing for
     the specific event type */
  if (event_widget == widget)
    {
      return FALSE;
    }

  w = event_widget;
  while (w && w != widget)
    w = w->parent;

  if (w)			/* event was in child of grab widget */
    {
      /* This would more aproporiately be gtk_propagate_event from
	 gtkmain.c, but that isn't exported */
      gtk_widget_event(event_widget,event);
    }

  return TRUE;
}
