/* GDK - The GIMP Drawing Kit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __GDK_INPUT_H__
#define __GDK_INPUT_H__

/* FIXME this will _break_ */

#include <X11/extensions/XInput.h>

typedef struct _GdkAxisInfo    GdkAxisInfo;
typedef struct _GdkInputVTable GdkInputVTable;
typedef struct _GdkDevicePrivate GdkDevicePrivate;
typedef struct _GdkInputWindow GdkInputWindow;

struct _GdkInputVTable {
  gint (*set_mode) (guint32 deviceid, GdkInputMode mode);
  void (*set_axes) (guint32 deviceid, GdkAxisUse *axes);
  GdkDeviceTimeCoord* (*motion_events) (guint32 deviceid,
					guint32 start,
					guint32 stop,
					gint *nevents_return);
  void (*get_pointer)   (GdkWindow       *window,
			 guint32	  deviceid,
			 gdouble         *x,
			 gdouble         *y,
			 gdouble         *pressure,
			 gdouble         *xtilt,
			 gdouble         *ytilt,
			 GdkModifierType *mask);

  /* Handle an unidentified event. Returns TRUE if handled, FALSE
     otherwise */
  void (*configure_event) (XConfigureEvent *xevent, GdkWindow *window);
  void (*enter_event) (XCrossingEvent *xevent, GdkWindow *window);
  gint (*other_event) (GdkEvent *event, XEvent *xevent, GdkWindow *window);
  gint (*window_none_event) (GdkEvent *event, XEvent *xevent);
  gint (*enable_window) (GdkWindow *window, GdkDevicePrivate *gdkdev);
  gint (*disable_window) (GdkWindow *window, GdkDevicePrivate *gdkdev);
};

/* information about a device axis */
struct _GdkAxisInfo
{
  /* reported x resolution */
  gint xresolution;

  /* reported x minimum/maximum values */
  gint xmin_value, xmax_value;

  /* calibrated resolution (for aspect ration) - only relative values
     between axes used */
  gint resolution;
  
  /* calibrated minimum/maximum values */
  gint min_value, max_value;
};

#define GDK_INPUT_NUM_EVENTC 6

struct _GdkDevicePrivate {
  GdkDeviceInfo  info;

  /* information about the axes */
  GdkAxisInfo *axes;

  /* reverse lookup on axis use type */
  gint axis_for_use[GDK_AXIS_LAST];
  
  /* Information about XInput device */
  XDevice       *xdevice;

  int buttonpress_type, buttonrelease_type, motionnotify_type,
      proximityin_type, proximityout_type, changenotify_type;

  /* true if we need to select a different set of events, but
     can't because this is the core pointer */
  gint needs_update;

  /* Mask of buttons (used for button grabs) */
  gint button_state;

  /* true if we've claimed the device as active. (used only for XINPUT_GXI) */
  gint claimed;
};

struct _GdkInputWindow
{
  /* gdk window */
  GdkWindow *window;

  /* position relative to root window */
  gint16 root_x;
  gint16 root_y;

  /* rectangles relative to window of windows obscuring this one */
  GdkRectangle *obscuring;
  gint num_obscuring;
};

/* Global data */

extern GdkInputVTable gdk_input_vtable;
extern GdkDevicePrivate *gdk_input_core_pointer;
extern GdkDevicePrivate *gdk_input_current_device;
extern GList           *gdk_input_devices;
extern GList           *gdk_input_windows;
/* information about network port and host for gxid daemon */
extern gchar           *gdk_input_gxid_host;
extern gint             gdk_input_gxid_port;
extern gint             gdk_input_ignore_core;

/* Function declarations */

void gdk_input_window_destroy (GdkWindow *window);

#endif __GDK_INPUT_H__
