#include <glib/glib.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fnmatch.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <gdk/gdkprivate.h>

#include <gdk/automata.h>

int hangul_debug = 0;
int use_hangul = 1; /* loose */

/* currently code and font are supported wansung only. */
int
convert_code_to_font(unsigned char *src, XChar2b *buf, int n)
{
    XChar2b *buf_orig = buf;
    while(n>0) {
	buf->byte1 = (*src++) & 0x7f;
	buf++->byte2 = (*src++) & 0x7f;
	n -= 2;
    }
    return (buf - buf_orig);
}

XFontStruct *
hangul_font_find(XFontStruct *eng_font)
{
    static int initialize = 0;
    static GSList *english_font_list = NULL;
    static GSList *hangul_font_list = NULL;
    GSList *elist,*hlist;
	XFontStruct *hangul_font;
    char *eng_full_name = NULL;
    XFontProp *pp;
	Atom atom_FONT;
    int i;
	if(eng_font == NULL) {
		g_print("english font is (nil) in %s\n", __FUNCTION__);
		return NULL;
	}
    atom_FONT = XInternAtom(gdk_display, "FONT", False);
    for(i=0, pp = eng_font->properties; i< eng_font->n_properties;i++,pp++)
    	if(pp->name == atom_FONT) {
	    eng_full_name = XGetAtomName(gdk_display, pp->card32);
	    break;
	}
    if(eng_full_name == NULL) {
	g_warning("can't find fullname of english font \n");
	return NULL;
    }
    for(i=0;eng_full_name[i];i++) {
	eng_full_name[i] = tolower(eng_full_name[i]);
    }
    if(hangul_debug) {
	fprintf(stderr,"english font name:: '%s' ==>\n",eng_full_name);
    }

    if(!initialize) {
	char config_file[256];
	FILE *fp;
	int line_number = 0;
	char *s;
	char buf[1024];
	sprintf(config_file,"%s/.hgdkrc", getenv("HOME"));
	fp = fopen(config_file,"r");
	if(fp == NULL) {
	    g_warning("~/.hgdkrc file open failure\n");
	    return NULL;
	}
	while(1) {
		extern int hangul_keyboard_type;
	    line_number++;
	    if(fgets(buf, 1024, fp) == NULL) break;
	    if(*buf == '#') continue;
	    if((hangul_keyboard_type == 0) && strncmp(buf,"keyboard=", 9) == 0) {
		int keyboard = atoi(buf + 9);
		if(keyboard != 2 && keyboard != 3) {
		    fprintf(stderr,"error in parsing %s: \
		       invalid keyboard type %d\n", config_file, keyboard);
		} else 
		    automata_keyboard_set(keyboard);
		continue;
	    }
	    s = strchr(buf, '=');
	    if(s == NULL) {
		fprintf(stderr,"error in parsing %s: in line %d\n",config_file,
		 line_number);
		 continue;
	    }
	    *s++ = 0;
	    if(s[strlen(s) - 1] == '\n') s[strlen(s)-1] = 0;

	    english_font_list = 
	    	g_slist_append(english_font_list, g_strdup(buf));
	    hangul_font_list =
	    	g_slist_append(hangul_font_list, g_strdup(s));
	}
	initialize = 1;
    }
    elist = english_font_list;
    hlist = hangul_font_list;
    while(elist) {
	if(fnmatch(elist->data, eng_full_name,0) == 0) {
		if(*((char *)hlist->data) != 0){ 
			hangul_font = XLoadQueryFont(gdk_display, (char *)hlist->data);
			if(hangul_font == NULL) {
				g_warning("could not load hangul font(%s)\n", (char *)hlist->data);
			}
		} else 
			hangul_font = NULL;
		return hangul_font;
	}
	elist = elist->next;
	hlist = hlist->next;
    }
    g_warning("could not find the hangul font matching\n");
	g_warning("%s \n", eng_full_name);
    return NULL;
}

int
HTextWidth(XFontStruct *english_font, XFontStruct *hangul_font, 
	char *s, int len)
{
    int hfont_width = hangul_font->max_bounds.width/2;
    char *p;
    int width = 0;

    while(len>0) {
	if(*s & 0x80) {
	    for(p=s;(len > 0) && (*p & 0x80); p+=2,len-=2) ;
	    width += hfont_width * (p -s);
	    s = p;
	}
	for(p=s;(len>0) && !(*p & 0x80); p++,len--);
	width += XTextWidth(english_font, s, (p-s));
	s = p;
    }
    return width;
}

void
HDrawString(d, w, gc, hgc, font, x, y, str, len, hlift)
    Display *d;
    Window w;
    GC gc;
    GC hgc;
    GdkFontPrivate *font;
    int x, y;
    char* str;
    int len,hlift;
{
    XChar2b buf[200];
    int n;
    char *p;
    XFontStruct *english_font = font->xfont;
    int hangul_font_width = font->hfont->max_bounds.width;

    while (len > 0) {
	if (*str & 0x80) {
	    for (p = str; len > 0 && (*p & 0x80) ; p += 2, len -= 2) ;
	    n = convert_code_to_font((unsigned char *)str, buf, (int)(p - str));
	    if(n) {
		XDrawString16(d, w, hgc, x, y -hlift, buf, n);
		x +=  hangul_font_width * n;
		str = p;
	    }
	}
	for (p = str; len > 0 && !(*p & 0x80); p++, len--) {
	}
	XDrawString(d, w, gc, x, y, str, p - str);
	x += XTextWidth(english_font, str, (p-str));
	str = p;
    }
}

int
is_first_han(char *buf, int x)
{
	int xx = x;
	if(*(buf + x) & 0x80) {
		do { 
			x--;
		} while(x >= 0 && *(buf + x) & 0x80);
		if((xx - x)%2 == 1) return TRUE;
	} 
	return FALSE;
}

int
is_second_han(char *buf, int x)
{
	int xx = x;
	if(*(buf + x) & 0x80) {
		do { 
			x--;
		} while(x >= 0 && *(buf + x) & 0x80);
		if((xx - x)%2 == 0) return TRUE;
	} 
	return FALSE;
}

